from pywapor.et_look_dev import constants as c
import numpy as np
import xarray as xr

def evaporative_fraction(et_24_mm, lh_24, rn_24, g0_24):
    r"""
    Computes the evaporative fraction.

    where the following constants are used

    * :math:`d_{sec}` seconds in the day = 86400 [s]

    Parameters
    ----------
    et_24_mm : float
        daily actual evapotranspiration
        :math:`ET_{ref}`
        [mm]
    lh_24 : float
        daily latent heat of evaporation
        :math:`\lambda_{24}`
        [J/kg]
    rn_24 : float
        daily net radiation
        :math:`Q^{*}`
        [Wm-2]
    g0_24 : float
        daily soil heat flux
        :math:`G`
        [W m-2]
        
    Returns
    -------
    ef_24 : float
        evaporative fraction
        :math:`EF_{ref}`
        [-]
    """
    et_24 = et_24_mm * lh_24/c.day_sec
    ef_24 = et_24/(rn_24 - g0_24)
    
    return ef_24


def interception_mm(P_24, vc, lai, int_max=0.2):
    r"""
    Computes the daily interception. The daily interception of a vegetated area
    is calculated according to von Hoyningen-Hüne (1983) [Ho1983]_
    and Braden(1985) [Br1985]_.

    .. math ::
        I^*=I_{max}*I_{lai}*\left(1-\left(\frac{1}{1+\frac{c_{veg}P24}
        {I_{max}I_{lai}}}\right)\right)

    Parameters
    ----------
    P_24 :  float
        daily rainfall
        :math:`P`
        [mm day :math:`^{-1}`]

    vc : float
        vegetation cover
        :math:`c_{veg}`
        [-]

    lai : float
        leaf area index
        :math:`I_{lai}`
        [-]

    int_max :  float
        maximum interception per leaf
        :math:`I_{max}`
        [mm day :math:`^{-1}`]

    Returns
    -------
    int_mm : float
        interception
        :math:`I^*`
        [mm day :math:`^{-1}`]

    References
    ----------
    .. [Br1985] Braden, H., Energiehaushalts- und Verdunstungsmodell für Wasser- und
        Stoffhaushalts-untersuchungen landwirtschaftlich genutzter
        Einzugsgebiete. Mitteilungen der Deutschen Bodenkundlichen
        Gesellschaft, (1985), 42, 254-299
    .. [Ho1983] von Hoyningen-Hüne, J., Die Interception des Niederschlags in
        landwirtschaftlichen Beständen. Schriftenreihe des DVWK, 1983, 57, 1-53


    """
    zero_mask = np.logical_or.reduce((lai == 0, vc == 0, P_24 == 0))

    res = int_max * lai * (1 - (1 / (1 + ((vc * P_24) / (int_max * lai)))))

    if isinstance(res, xr.DataArray):
        res = xr.where(zero_mask, 0.0, res)
    else:
        res[zero_mask] = 0

    return res


def et_reference(rn_24_grass, ad_24, psy_24, vpd_24, ssvp_24, u_24):
    r"""
    Computes the reference evapotranspiration. The reference evapotranspiration
    :math:`ET_{ref}` is an important concept in irrigation science. The reference
    evapotranspiration can be inferred from routine meteorological
    measurements. The reference evapotranspiration is the evapotranspiration
    of grass under well watered conditions.
    First the aerodynamical resistance for grass :math:`r_{a,grass}` [sm :math:`^{-1}`]
    is calculated

    .. math ::

        r_{a,grass}=\frac{208}{u_{obs}}

    Then the reference evapotranspiration :math:`ET_{ref}` [W m :math:`^{-2}`] can be calculated
    as follows, with taking the default value for the grass surface resistance
    :math:`r_{grass}` = 70 sm :math:`^{-1}`

    .. math ::
        ET_{ref}=\frac{\Delta\left(Q_{grass}^{*}\right)+
        \rho c_{p}\frac{\Delta_{e}}{r_{a,grass}}}
        {\Delta+\gamma\left(1+\frac{r_{grass}}{r_{a,grass}}\right)}

    The soil heat flux is assumed to be zero or close to zero on a daily basis.

    Parameters
    ----------
    rn_24_grass : float
        net radiation for reference grass surface
        :math:`Q^{*}_{grass}`
        [Wm-2]
    u_24 : float
        daily wind speed at observation height
        :math:`u_{obs}`
        [m/s]
    ad_24 : float
        daily air density
        :math:`\rho_{24}`
        [kg m-3]
    psy_24 : float
        daily psychrometric constant
        :math:`\gamma_{24}`
        [mbar K-1]
    vpd_24 : float
        daily vapour pressure deficit
        :math:`\Delta_{e,24}`
        [mbar]
    ssvp_24 : float
        daily slope of saturated vapour pressure curve
        :math:`\Delta_{24}`
        [mbar K-1]

    Returns
    -------
    et_ref_24 : float
        reference evapotranspiration (well watered grass) energy equivalent
        :math:`ET_{ref}`
        [W m-2]
    """
    r_grass = 70
    ra_grass = 208. / u_24
    et_ref_24 = np.maximum(0,(ssvp_24 * rn_24_grass + ad_24 * c.sh * (vpd_24 / ra_grass)) /\
        (ssvp_24 + psy_24 * (1 + r_grass / ra_grass)))
    return et_ref_24


def et_reference_mm(et_ref_24, lh_24):
    r"""
    Computes the reference evapotranspiration.

    .. math ::

        ET_{ref}=ET_{ref}d_{sec}\lambda_{24}

    where the following constants are used

    * :math:`d_{sec}` seconds in the day = 86400 [s]

    Parameters
    ----------
    et_ref_24 : float
        daily reference evapotranspiration energy equivalent
        :math:`ET_{ref}`
        [W m-2]
    lh_24 : float
        daily latent heat of evaporation
        :math:`\lambda_{24}`
        [J/kg]

    Returns
    -------
    et_ref_24_mm : float
        reference evapotranspiration (well watered grass)
        :math:`ET_{ref}`
        [mm d-1]
    """
    return et_ref_24 * c.day_sec / lh_24


def et_actual_mm(e_24_mm, t_24_mm):
    r"""
    Computes the actual evapotranspiration based on the separate calculations
    of evaporation and transpiration:

    .. math ::
        ET = E + T

    Parameters
    ----------
    e_24_mm : float
        daily evaporation in mm
        :math:`E`
        [mm d-1]
    t_24_mm : float
        daily transpiration in mm
        :math:`T`
        [mm d-1]

    Returns
    -------
    et_24_mm : float
        daily evapotranspiration in mm
        :math:`ET`
        [mm d-1]
    """
    return e_24_mm + t_24_mm



import xarray as xr
import numpy as np
import os
import numpy as np
import matplotlib.pyplot as plt
import xarray as xr
from pywapor.enhancers.smooth.whittaker import whittaker, cross_val_lmbda, second_order_diff_matrix, cve1, cve2, whittaker_smoothing
from pywapor.enhancers.smooth.core import wt1, wt2
import pandas as pd
import os
import itertools
import pywapor

def filter_ts(x, y, u = None, tol = 0.002, chunks = "auto", sample_interval = 1, xdim = "time"):

    if not isinstance(u, type(None)):
        if not isinstance(u, xr.DataArray):
            u = xr.DataArray(u, dims = x.dims, coords = x.coords)

    if sample_interval > 1:
        x = x.isel({xdim: slice(None, None, sample_interval)})
        y = y.isel({xdim: slice(None, None, sample_interval)})

    diff_forward = x.diff(xdim).assign_coords({xdim: x.time[:-1]})
    diff_backward = x.diff(xdim).assign_coords({xdim: x.time[1:]})
    d2x = (diff_forward + diff_backward).reindex_like(x.time, fill_value = tol * 2)
    
    nremoving = (d2x < tol).sum().values
    if nremoving > 0:
        log.info(f"--> Removing {nremoving} {xdim} slices.")

    if not isinstance(u, type(None)):
        if sample_interval > 1:
            u = u.isel({xdim: slice(None, None, sample_interval)})
        u = u.where(d2x >= tol, drop = True)
        u = u.values

    y = y.where(d2x >= tol, drop = True).chunk(chunks)
    x = x.where(d2x >= tol, drop = True)

    return x, y, u

def open_ts(fps):

    dss = list()
    for i, fp in enumerate(fps):

        name = os.path.split(fp)[-1].replace(".nc", "")
        ds = xr.open_dataset(fp, decode_coords = "all")

        if i > 0:
            if ds.x.size != example_ds.x.size or ds.y.size != example_ds.y.size:
                ds = ds.rio.reproject_match(example_ds)

        ds["sensor"] = xr.ones_like(ds["time"], dtype = int).where(False, name)
        dss.append(ds)

        if i ==0:
            example_ds = ds.copy()
    
    ds = xr.concat(dss, dim = "time", combine_attrs = "drop_conflicts").sortby("time").transpose("y", "x", "time")
    ds["sensor"] = ds["sensor"].astype("<U7")
    ds.attrs = {}

    attribute = {str(i): sensor_name for i, sensor_name in enumerate(np.unique(ds.sensor))}
    values = np.array(list(attribute.keys()), dtype = int)
    coords = np.array(list(attribute.values()), dtype = str)
    transformer = xr.DataArray(values, dims=["sensor"], coords = {"sensor": coords})
    ds["sensor"] = transformer.sel(sensor = ds.sensor).drop("sensor").assign_attrs(attribute)

    return ds

def cross_val_lmbda(y, x, lmbdas = np.logspace(0, 3, 4), 
                    u = None, chunks = "auto", tol = 0.02,
                    sample_interval = 1):

    x, y, lmbdas, dim_name = assert_dtypes(x, y, lmbdas)

    # Normalize x-coordinates
    x = (x - x.min()) / (x.max() - x.min()) * x.size

    # Remove points that are too close together
    if isinstance(x, xr.DataArray):
        x, y, u = filter_ts(x, y, u = u, tol = tol, chunks = chunks, 
                            sample_interval = sample_interval, xdim = dim_name)

    # Create x-aware delta matrix.
    D = second_order_diff_matrix(x)
    A = np.dot(D.T, D)

    # Choose which function to use depending on shapes of y and lmbdas.
    # cve = choose_func(y, lmbdas, "cve")

    if isinstance(u, type(None)):
        u = np.ones_like(x)

    if isinstance(x, xr.DataArray):

        # # Determine dimension names.
        # if cve.__name__ == "cve1":
        #     icd = [[], [dim_name], [], []]
        #     ocd = [[]]
        # elif cve.__name__ == "cve2":
        #     icd = [["lmbda"], [dim_name], [], []]
        #     ocd = [["lmbda"]]

        # Make sure lmbdas is chunked similar to y.
        if not isinstance(y.chunk, type(None)):
            lmbdas = lmbdas.chunk({k: v for k,v in y.unify_chunks().chunksizes.items() if k in lmbdas.dims})

        # NOTE this is done to supress a warning ("invalid value encountered in cve2")
        # which is given when a column (along `dim_name`) doenst contain any finite value.
        # Its currently impossible to supress this warning, so filling those columns with
        # fake data (-99) here and then masking those pixels again afterwards.
        # See https://github.com/dask/dask/issues/3245 for more info.
        if cve.__name__ == "cve2":
            mask = y.count(dim_name) > 0
            y = y.where(mask, -99)

        # Calculate the cross validation standard error for each lambda using
        # the hat matrix.
        cves = xr.apply_ufunc(
            cve, lmbdas, y, A, u,
            input_core_dims=icd,
            output_core_dims=ocd,
            dask = "allowed")

        # NOTE see above.
        if cve.__name__ == "cve2":
            cves = cves.where(mask, np.nan)

        # Select the lambda for which the error is smallest.
        if "lmbda" in cves.dims:
            lmbda_sel = cves.idxmin(dim = "lmbda")
            lmbda_sel = lmbda_sel.fillna(lmbdas.min().values)
        else:
            cves.assign_coords({"lmbda": lmbdas})
            lmbda_sel = lmbdas
    else:
        cves = cve(lmbdas, y, A, u)
        if np.isscalar(lmbdas):
            lmbda_sel = lmbdas
        elif lmbdas.ndim == 1:
            idx = np.argmin(cves, axis = -1)
            lmbda_sel = lmbdas[idx]
        else:
            lmbda_sel = lmbdas
        
    return lmbda_sel, cves

def cross_val_lmbda_ref(y, x, lmbdas = np.logspace(-2, 3, 100)):
    # Determine vector length.
    m = len(x)
    # Create arrays of m timeseries in which each time one value has been removed.
    Y_ = np.where(np.eye(m), np.nan, np.repeat(y[np.newaxis, :], m, axis = 0))
    # Smooth the timeeseries and make an estimate for the removed value for each lambda.
    z = whittaker(Y_, x, lmbdas = lmbdas)
    # Retrieve estimated timeseries for each lambda.
    y_hat = np.diagonal(z, axis1 = 0, axis2 = 2)
    # Calculate the cross validation standard error for each lambda.
    cves = np.sqrt(np.nanmean(((y - y_hat)**2), axis = 1))
    # Select the lambda for which the error is smallest.
    lmbda_sel = lmbdas[np.argmin(cves)]
    return lmbda_sel, cves

def test_cross_val(y, x):
    lmbdas = np.logspace(-2, 2, 20)
    lmbda_sel1, cves1 = cross_val_lmbda_ref(y, x, lmbdas = lmbdas)
    lmbda_sel2, cves2 = cross_val_lmbda(y, x, lmbdas = lmbdas)
    # assert np.isclose(lmbda_sel1, lmbda_sel2)
    plt.plot(lmbdas, cves1, label = "1")
    plt.plot(lmbdas, cves2, label = "2")
    plt.legend()

def test_shapes_z(y, x, xn = 2, yn = 5, n = 4, m = 3, a = 0.5, max_iter = 10):
    t = len(y)

    Y1 = y[:]
    Y2 = np.repeat(y, xn*yn).reshape((xn,yn,t), order = "F")
    Y3 = np.repeat(y, n).reshape((n,t), order = "F")

    LMB1 = 100.
    LMB2 = np.logspace(1,3,m)
    LMB3 = np.logspace(1,3,xn*yn).reshape((xn, yn))

    # Normalize x-coordinates
    x = (x - x.min()) / (x.max() - x.min()) * x.size

    # Create x-aware delta matrix.
    D = second_order_diff_matrix(x)
    A = np.dot(D.T, D)

    u = np.ones_like(x)

    z11 = wt1(Y1, A, LMB1, u, a, -np.inf, np.inf, max_iter)
    assert z11.shape == (t,)

    z21 = wt1(Y2, A, LMB1, u, a, -np.inf, np.inf, max_iter)
    assert z21.shape == (xn,yn,t)

    z31 = wt1(Y3, A, LMB1, u, a, -np.inf, np.inf, max_iter)
    assert z31.shape == (n,t)

    z12 = wt1(Y1, A, LMB2, u, a, -np.inf, np.inf, max_iter) #
    assert z12.shape == (m,t)

    z22 = wt2(Y2, A, LMB2, u, a, -np.inf, np.inf, max_iter) #
    assert z22.shape == (xn,yn,m,t)

    z32 = wt2(Y3, A, LMB2, u, a, -np.inf, np.inf, max_iter) #
    assert z32.shape == (n,m,t)

    z13 = wt1(Y1, A, LMB3, u, a, -np.inf, np.inf, max_iter) #
    assert z13.shape == (xn,yn,t)

    z23 = wt1(Y2, A, LMB3, u, a, -np.inf, np.inf, max_iter) #!
    assert z23.shape == (xn,yn,t)

def test_shapes_cve(y, x, xn = 2, yn = 5, n = 4, m = 3):
    t = len(y)

    Y1 = y[:]
    Y2 = np.repeat(y, xn*yn).reshape((xn,yn,t), order = "F")
    Y3 = np.repeat(y, n).reshape((n,t), order = "F")

    LMB1 = 100.
    LMB2 = np.logspace(1,3,m)
    LMB3 = np.logspace(1,3,xn*yn).reshape((xn, yn))
    
    # Normalize x-coordinates
    x = (x - x.min()) / (x.max() - x.min()) * x.size

    # Create x-aware delta matrix.
    D = second_order_diff_matrix(x)
    A = np.dot(D.T, D)

    u = np.ones_like(x)

    # cve11 = cve1(LMB1, Y1, A, u)
    # assert cve11.shape == ()

    # cve21 = cve1(LMB1, Y2, A, u)
    # assert cve21.shape == (xn,yn)

    # cve31 = cve1(LMB1, Y3, A, u)
    # assert cve31.shape == (n,)

    cve12 = cve1(LMB2, Y1, A, u)
    assert cve12.shape == (m,)

    cve22 = cve2(LMB2, Y2, A, u)
    assert cve22.shape == (xn,yn,m)

    cve32 = cve2(LMB2, Y3, A, u)
    assert cve32.shape == (n,m)

    cve13 = cve1(LMB3, Y1, A, u)
    assert cve13.shape == (xn,yn)

    cve23 = cve1(LMB3, Y2, A, u)
    assert cve23.shape == (xn,yn)

def test_whittaker_main(ds, base = 10, xn = 2, yn = 5, n = 4, m = 3):

    t = len(ds.time)
    X = ds.time

    Y1 = ds.isel(y = base, x = base)["ndvi"]
    Y2 = ds.isel(y = slice(base, base + yn), x = slice(base, base + xn))["ndvi"].transpose("y","x","time")
    Y3 = ds.isel(y = slice(base, base + n), x = base).rename({"y": "n"})["ndvi"]

    u = np.ones((t)) * 0.5

    LMB1 = xr.DataArray(100.)
    LMB2 = xr.DataArray(np.logspace(1,3,m), dims = ["lmbda"], coords = {"lmbda": np.logspace(1,3,m)})
    LMB3 = xr.DataArray(np.logspace(1,3,xn*yn).reshape((xn, yn)), dims = ["x", "y"], coords = {"x": Y2.x, "y": Y2.y}).transpose("y","x")
    
    out1 = whittaker(Y1, X, lmbdas = LMB1, u = u)
    assert isinstance(out1, xr.DataArray)
    assert np.all([v == {"time": t}[k] for k, v in out1.sizes.items()])

    out2 = whittaker(Y2, X, lmbdas = LMB1, u = u)
    assert isinstance(out2, xr.DataArray)
    assert np.all([v == {"x": xn, "y": yn, "time": t}[k] for k, v in out2.sizes.items()])

    out3 = whittaker(Y3, X, lmbdas = LMB1, u = u)
    assert isinstance(out3, xr.DataArray)
    assert np.all([v == {"n": n, "time": t}[k] for k, v in out3.sizes.items()])

    out4 = whittaker(Y1, X, lmbdas = LMB2, u = u)
    assert isinstance(out4, xr.DataArray)
    assert np.all([v == {"lmbda": m, "time": t}[k] for k, v in out4.sizes.items()])

    out5 = whittaker(Y2, X, lmbdas = LMB2, u = u)
    assert isinstance(out5, xr.DataArray)
    assert np.all([v == {"lmbda": m, "time": t, "x": xn, "y": yn}[k] for k, v in out5.sizes.items()])

    out6 = whittaker(Y3, X, lmbdas = LMB2, u = u)
    assert isinstance(out6, xr.DataArray)
    assert np.all([v == {"lmbda": m, "time": t, "n": n}[k] for k, v in out6.sizes.items()])

    out7 = whittaker(Y1, X, lmbdas = LMB3, u = u)
    assert isinstance(out7, xr.DataArray)
    assert np.all([v == {"x": xn, "time": t, "y": yn}[k] for k, v in out7.sizes.items()])

    out8 = whittaker(Y2, X, lmbdas = LMB3, u = u)
    assert isinstance(out8, xr.DataArray)
    assert np.all([v == {"x": xn, "time": t, "y": yn}[k] for k, v in out8.sizes.items()])

    ######

    out1 = whittaker(Y1.values, X.values, lmbdas = LMB1.values, u = u)
    assert out1.shape == (t,)

    out2 = whittaker(Y2.values, X.values, lmbdas = LMB1.values, u = u)
    assert out2.shape == (yn,xn,t)

    out3 = whittaker(Y3.values, X.values, lmbdas = LMB1.values, u = u)
    assert out3.shape == (n,t)

    out4 = whittaker(Y1.values, X.values, lmbdas = LMB2.values, u = u)
    assert out4.shape == (m,t)

    out5 = whittaker(Y2.values, X.values, lmbdas = LMB2.values, u = u)
    assert out5.shape == (yn,xn,m,t)

    out6 = whittaker(Y3.values, X.values, lmbdas = LMB2.values, u = u)
    assert out6.shape == (n,m, t)

    out7 = whittaker(Y1.values, X.values, lmbdas = LMB3.values, u = u)
    assert out7.shape == (yn,xn,t)

    out8 = whittaker(Y2.values, X.values, lmbdas = LMB3.transpose("y","x").values, u = u)
    assert out8.shape == (yn,xn,t)

def test_cve_main(ds, base = 10, xn = 2, yn = 5, n = 4, m = 3):

    X = ds.time

    Y1 = ds.isel(y = base, x = base)["ndvi"]
    Y2 = ds.isel(y = slice(base, base + yn), x = slice(base, base + xn))["ndvi"].transpose("y","x","time")
    Y3 = ds.isel(y = slice(base, base + n), x = base).rename({"y": "n"})["ndvi"]

    u = np.ones((X.size)) * 0.5

    LMB1 = xr.DataArray(100.)
    LMB2 = xr.DataArray(np.logspace(1,3,m), dims = ["lmbda"], coords = {"lmbda": np.logspace(1,3,m)})
    LMB3 = xr.DataArray(np.logspace(1,3,xn*yn).reshape((xn, yn)), dims = ["x", "y"], coords = {"x": Y2.x, "y": Y2.y}).transpose("y","x")
    
    out1, cves1 = cross_val_lmbda(Y1, X, lmbdas = LMB1, u = u)
    assert out1 == LMB1
    assert cves1.ndim == 0

    out2, cves2 = cross_val_lmbda(Y2, X, lmbdas = LMB1, u = u)
    assert out2 == LMB1
    assert np.all([v == {"x": xn, "y": yn,}[k] for k, v in cves2.sizes.items()])

    out3, cves3 = cross_val_lmbda(Y3, X, lmbdas = LMB1, u = u)
    assert out3 == LMB1
    assert np.all([v == {"n": n}[k] for k, v in cves3.sizes.items()])

    out4, cves4 = cross_val_lmbda(Y1, X, lmbdas = LMB2, u = u)
    assert np.all([v == {"lmbda": m}[k] for k, v in cves4.sizes.items()])
    assert out4.ndim == 0

    out5, cves5 = cross_val_lmbda(Y2, X, lmbdas = LMB2, u = u)
    assert np.all([v == {"lmbda": m, "x": xn, "y": yn}[k] for k, v in cves5.sizes.items()])
    assert np.all([v == {"x": xn, "y": yn}[k] for k, v in out5.sizes.items()])

    out6, cves6 = cross_val_lmbda(Y3, X, lmbdas = LMB2, u = u)
    assert np.all([v == {"lmbda": m, "n": n}[k] for k, v in cves6.sizes.items()])
    assert np.all([v == {"n": n}[k] for k, v in out6.sizes.items()])

    out7, cves7 = cross_val_lmbda(Y1, X, lmbdas = LMB3, u = u)
    assert out7.equals(LMB3)
    assert np.all([v == {"x": xn, "y": yn}[k] for k, v in cves7.sizes.items()])

    out8, cves8 = cross_val_lmbda(Y2, X, lmbdas = LMB3, u = u)
    assert out8.equals(LMB3)
    assert np.all([v == {"x": xn, "y": yn}[k] for k, v in cves8.sizes.items()])

    ######

    out1, cves1 = cross_val_lmbda(Y1.values, X.values, lmbdas = LMB1.values, u = u)
    assert out1 == LMB1.values
    assert np.isscalar(cves1)

    out2, cves2 = cross_val_lmbda(Y2.values, X.values, lmbdas = LMB1.values, u = u)
    assert out3 == LMB1.values
    assert cves2.shape == (yn,xn)

    out3, cves3 = cross_val_lmbda(Y3.values, X.values, lmbdas = LMB1.values, u = u)
    assert out3 == LMB1.values
    assert cves3.shape == (n,)

    out4, cves4 = cross_val_lmbda(Y1.values, X.values, lmbdas = LMB2.values, u = u)
    assert np.isscalar(out4)
    assert cves4.shape == (m,) 

    out5, cves5 = cross_val_lmbda(Y2.values, X.values, lmbdas = LMB2.values, u = u)
    assert out5.shape == (yn, xn)
    assert cves5.shape == (yn,xn,m) 

    out6, cves6 = cross_val_lmbda(Y3.values, X.values, lmbdas = LMB2.values, u = u)
    assert out6.shape == (n,)
    assert cves6.shape == (n, m)  

    out7, cves7 = cross_val_lmbda(Y1.values, X.values, lmbdas = LMB3.values, u = u)
    assert np.all(out7 == LMB3.values)
    assert cves7.shape == (yn, xn)  

    out8, cves8 = cross_val_lmbda(Y2.values, X.values, lmbdas = LMB3.values, u = u)
    assert np.all(out8 == LMB3.values)
    assert cves8.shape == (yn, xn)  

def test_whittaker_smoothing(ds, var, m = 3):

    xn = ds.x.size
    yn = ds.y.size

    LMB0 = 100.
    LMB1 = xr.DataArray(100.)
    LMB2 = xr.DataArray(np.logspace(1,3,m), dims = ["lmbda"], coords = {"lmbda": np.logspace(1,3,m)})
    LMB3 = xr.DataArray(np.logspace(1,3,xn*yn).reshape((xn, yn)), dims = ["x", "y"], coords = {"x": ds.x, "y": ds.y}).transpose("y","x")
    LMB4 = LMB1.values
    LMB5 = LMB2.values

    _LMB = [
            LMB0, 
            LMB1, LMB2, 
            LMB3, LMB4, LMB5
            ]

    out_fh = None
    xdim = "time"
    new_x1 = None
    new_x2 = pd.date_range(np.datetime_as_string(ds.time[0], "D") + " 12:00", 
                           np.datetime_as_string(ds.time[-1], "D") + " 12:00", freq="D")
    _new_x = [new_x1, new_x2]
    _export_all = [True, False]
    _DS = [
        ds.load(), 
        ds.chunk({"x": 5, "y": 5, "time": -1})
        ]

    _a = [0.5, 0.1]

    _u = [None, {"LC08_SR": 1.0, "LC09_SR": 1.0, "LE07_SR": 0.7}]

    tests = list(itertools.product(*[_DS, _LMB, _new_x, _export_all, _a, _u]))

    files = list()

    for i, (DS, LMB, new_x, export_all, a, u) in enumerate(tests):

        print(f"TEST {i:>03} / {len(tests):>03}")
        print("ds", DS)
        print("lmbda", LMB)
        print("new_x", new_x)
        print(export_all)
        print(a)
        print(u)
        print("------\n")

        out = whittaker_smoothing(DS, var,
                                lmbdas = LMB,
                                weights = u,
                                a = a,
                                max_iter = 10,
                                out_fh = out_fh,
                                xdim = xdim,
                                new_x = new_x,
                                export_all = export_all)
        
        out = out.compute(scheduler = "synchronous")

        if export_all:
            assert np.all(np.sort(list(out.data_vars)) == np.sort(["ndvi", "sensor", "lmbda_sel", "cves", "ndvi_smoothed"]))
            assert "lmbda" in out.coords
            assert out.time.size == ds.time.size + getattr(new_x, "size", 0)
        else:
            assert (out.time.size == new_x2.size) or (out.time.size == ds.time.size)
            assert list(out.data_vars) == ["ndvi_smoothed"]

        files.append(out_fh)
        out = out.close()
        # os.remove(fp)

    for fp in files:
        if not isinstance(fp, type(None)):
            if os.path.isfile(fp):
                os.remove(fp)

if __name__ == "__main__":

    ts_fp = os.path.join(pywapor.__path__[0], "enhancers", "smooth", "input_test_series.nc")

    ds = xr.open_dataset(ts_fp, decode_coords="all")
    var = "ndvi"
    y = ds[var]
    x = ds["time"]

    m = 3

    test_shapes_z(y.isel(y=10, x=10).values, x.values)
    test_shapes_cve(y.isel(y=10, x=10).values, x.values)
    # test_whittaker_main(ds)
    # test_cve_main(ds)
    # test_whittaker_smoothing(ds, var)


import os
from pywapor.collect import accounts
import datetime
import pandas as pd
import pywapor.collect.accounts as accounts
from pywapor.general.logger import log
import re
import warnings
from functools import partial
import requests
from pywapor.collect.protocol.crawler import download_urls, crawl
import pywapor.general.bitmasks as bm
import xarray as xr
from pywapor.enhancers.apply_enhancers import apply_enhancer
import numpy as np
from pywapor.general.processing_functions import process_ds, save_ds, open_ds, remove_ds
import datetime
import rioxarray.merge
import copy

def download(folder, latlim, lonlim, timelim, product_name, req_vars = ["ndvi", "r0"],
                variables = None, post_processors = None, timedelta = None):
    """Download MODIS data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list, optional
        Which variables to download for the selected product, by default ["ndvi", "r0"].
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """

    folder = os.path.join(folder, "PROBAV")

    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    dates = pd.date_range(timelim[0], timelim[1], freq="D")

    bb = (lonlim[0], latlim[0], lonlim[1], latlim[1])

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    coords = {"x": ["lon", None], "y": ["lat", None]}

    # TODO use urllib
    base_url = f"https://www.vito-eodata.be/PDF/datapool/Free_Data/PROBA-V_100m/{product_name}"
    coord_req = "?coord=" + ",".join([str(x) for x in bb])
    url = f"{base_url}/{coord_req}"

    session = requests.sessions.Session()
    session.auth = accounts.get("VITO")

    # Scrape urls.
    urls = find_tiles(url, dates, session)

    # Download .HDF5 tiles.
    fps = download_urls(urls, folder, session, parallel = 4)

    # Convert to netcdf.
    dss = dict()
    for fp in fps:
        date = datetime.datetime.strptime(fp.split("_")[-3], "%Y%m%d")
        ds = open_hdf5_groups(fp, variables, coords).expand_dims({"time": [date]})
        if date in dss.keys():
            dss[date].append(ds)
        else:
            dss[date] = [ds]

    # Merge tiles.
    dss0 = list()
    for date, datasets in dss.items():
        bb = (lonlim[0], latlim[0], lonlim[1], latlim[1])
        ds = rioxarray.merge.merge_datasets(datasets)
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category = FutureWarning)
            ds = ds.rio.clip_box(*bb) # NOTE using this seperately, because `bounds`-kw for `merge_datasets` bugs.
        dss0.append(ds)

    # Merge dates.
    ds = xr.concat(dss0, dim = "time")

    # Apply product specific functions.
    for var, funcs in post_processors.items():
        for func in funcs:
            ds, _ = apply_enhancer(ds, var, func)

    if product_name == "S5_TOC_100_m_C1":
        ds["time"] = ds["time"] + np.timedelta64(int(2.5 * 24), "h")

    ds = ds[req_vars]

    ds = save_ds(ds, fn, label = f"Merging files.")

    for k, v in dss.items():
        for nc in v:
            remove_ds(nc)

    return ds[req_vars_orig]

def open_hdf5_groups(fp, variables, coords):
    """Convert hdf5 with groups (specified in `variable`) to netCDF without groups.

    Parameters
    ----------
    fp : str
        Path to input file.
    variables : dict
        Metadata on which exact layers need to be requested from the server.
    coords : dict
        Metadata on the coordinates in `fp`.

    Returns
    -------
    xr.Dataset
        Ouput data.
    """

    nc_fp = fp.replace(".HDF5", ".nc")
    if os.path.isfile(nc_fp):
        ds = open_ds(nc_fp, "all")
    else:
        ds = xr.open_dataset(fp, chunks = "auto")

        spatial_ref_name = [k for k, v in variables.items() if v[1] == "spatial_ref"][0]
        if (spatial_ref_name in list(ds.variables)) and (spatial_ref_name not in list(ds.coords)):
            ds = ds.rio.write_grid_mapping(spatial_ref_name)
            ds = ds.set_coords((spatial_ref_name))

        for k in variables.keys():
            if k in ds.variables:
                continue
            with xr.open_dataset(fp, group = k, engine = "netcdf4", chunks = "auto") as ds_grp:
                vrs = list(ds_grp.variables)[0] # NOTE assuming each group has only 1 variable!
                ds[k] = ds_grp.rename({vrs: k})[k]

        ds = process_ds(ds, coords, variables)

        ds = save_ds(ds, nc_fp, label = f"Converting {os.path.split(fp)[-1]} to netcdf.")

    return ds

def mask_bitwise_qa(ds, var, flags):
    """Mask PROBAV data using a qa variable.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Variable in `ds` to mask.
    flags : list
        Which flags not to mask.

    Returns
    -------
    xr.Dataset
        Masked dataset.
    """
    flag_bits = bm.PROBAV_qa_translator()
    mask = bm.get_mask(ds["qa"].astype("uint8"), flags, flag_bits)
    ds[var] = ds[var].where(~mask, np.nan)
    return ds

def calc_r0(ds, *args):
    ds["r0"] = 0.429 * ds["blue"] + 0.333 * ds["red"] + 0.133 * ds["nir"] + 0.105 * ds["swir"]
    return ds

def default_post_processors(product_name, req_vars = ["ndvi", "r0"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["ndvi", "r0"].

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "S5_TOC_100_m_C1": {
            "r0": [
                    calc_r0, 
                    partial(mask_bitwise_qa, 
                    flags = ["bad BLUE", "bad RED", "bad NIR", "bad SWIR", 
                            "sea", "undefined", "cloud", "ice/snow", "shadow"]),
            ],
            "ndvi": [
                    partial(mask_bitwise_qa, 
                    flags = ["bad RED", "bad NIR", "sea", "undefined", 
                    "cloud", "ice/snow", "shadow"]),
            ],
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def default_vars(product_name, req_vars = ["ndvi", "r0"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["ndvi", "r0"].

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    
    variables = {
        "S5_TOC_100_m_C1": {
            "LEVEL3/NDVI":              [("lon", "lat"), "ndvi"],
            "LEVEL3/RADIOMETRY/BLUE":   [("lon", "lat"), "blue"],
            "LEVEL3/RADIOMETRY/NIR":    [("lon", "lat"), "nir"],
            "LEVEL3/RADIOMETRY/RED":    [("lon", "lat"), "red"],
            "LEVEL3/RADIOMETRY/SWIR":   [("lon", "lat"), "swir"],
            "LEVEL3/GEOMETRY/VNIR":     [("lon", "lat"), "vnir_vza"],
            "LEVEL3/GEOMETRY/SWIR":     [("lon", "lat"), "swir_vza"],
            "LEVEL3/QUALITY":           [("lon", "lat"), "qa"],
            "crs":                      [(), "spatial_ref"]
        },
    }

    req_dl_vars = {
        "S5_TOC_100_m_C1": {
            "ndvi": ["LEVEL3/NDVI", "LEVEL3/QUALITY", "crs"],
            "r0":   ["LEVEL3/RADIOMETRY/BLUE", "LEVEL3/RADIOMETRY/NIR", 
                    "LEVEL3/RADIOMETRY/RED", "LEVEL3/RADIOMETRY/SWIR",
                    "LEVEL3/QUALITY", "crs"],
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def find_tiles(url, dates, session):
    """Crawl webpage to find tile URLs.

    Parameters
    ----------
    url : str
        Path to start crawl.
    dates : pd.date_range
        Dates for which to return URLs.
    session : requests.Session
        Session to use during crawl.

    Returns
    -------
    list
        List with URLs to scenes.
    """

    log.info("--> Searching PROBAV tiles.")

    regex = "https:.*\/\d{4}\/\?coord="
    filter_regex = "\d{4}(?=\/\?coord=)"
    urls = {"_": url}
    label_filter = dates.strftime("%Y")
    years = crawl(urls, regex, filter_regex, session, label_filter = label_filter)

    regex = "https:.*\/\d{4}\/\d{2}\/\?coord="
    filter_regex = "\d{4}\/\d{2}(?=\/\?coord=)"
    label_filter = dates.strftime("%Y%m")
    months = crawl(years, regex, filter_regex, session, label_filter = label_filter)

    regex = "https:.*\/\d{4}\/\d{2}\/\d{2}\/\?coord="
    filter_regex = "\d{4}\/\d{2}\/\d{2}(?=\/\?coord=)"
    label_filter = dates.strftime("%Y%m%d")
    days = crawl(months, regex, filter_regex, session, label_filter = label_filter)

    regex = "https:.*\/\d{4}\/\d{2}\/\d{2}\/.*\/\?coord="
    filter_regex = "\d{4}\/\d{2}\/\d{2}(?=\/.*\/\?coord=)"
    label_filter = dates.strftime("%Y%m%d")
    prods = crawl(days, regex, filter_regex, session, label_filter = label_filter)

    regex = ".*\.HDF5"
    filter_regex = "\d{8}(?=.*\.HDF5)"
    fns = crawl(prods, regex, filter_regex, session, list_out = True)

    dl_urls = [os.path.join(re.sub("\?coord=.*","",prods[date_str]), fn) for date_str, fn, in fns]

    log.info(f"--> Downloading {len(dl_urls)} PROBAV tiles.")

    return dl_urls

if __name__ == "__main__":

    product_name = "S5_TOC_100_m_C1"

    folder = r"/Users/hmcoerver/Downloads/pywapor_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 2), datetime.date(2020, 7, 9)]

    variables = None
    post_processors = None
    req_vars = ["r0", "ndvi"]

    # ds = download(folder, latlim, lonlim, timelim, product_name, req_vars = req_vars,
    #             variables = variables, post_processors = post_processors)




"""
LXSS_LLLL_PPPRRR_YYYYMMDD_yyyymmdd_CX_TX
(e.g., LC08_L2SP_039037_20150728_20200318_02_T1)
L Landsat
X Sensor (“O” = OLI; “T” = TIRS; “C” = OLI/TIRS)
SS Satellite (“08” = Landsat 8, “09” = Landsat 9)
LLLL Processing correction level (“L2SP” if SR and ST are generated or “L2SR”
if ST could not be generated) PPP Path
RRR Row
YYYY Year of acquisition
MM Month of acquisition
DD Day of acquisition
yyyy Year of Level 2 processing
mm Month of Level 2 processing
dd Day of Level 2 processing
CX Collection number (“02”)
TX Collection category ( “T1” = Tier 1; “T2” = Tier 2)
"""

import requests as r
import time
import pywapor
import tarfile
import re
import shutil
import glob
import json
import warnings
import copy
import os
from functools import partial
import numpy as np
import xarray as xr
import datetime
import rasterio
from pywapor.general.logger import log, adjust_logger
from pywapor.general.processing_functions import open_ds, save_ds, remove_ds, adjust_timelim_dtype, make_example_ds
from pywapor.collect.protocol.crawler import download_urls
from pywapor.enhancers.apply_enhancers import apply_enhancer
from pywapor.enhancers.gap_fill import gap_fill
import xmltodict
from pywapor.general import bitmasks

def calc_r0(ds, var, product_name = None):
    """Calculate the Albedo.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.

    Returns
    -------
    xr.Dataset
        Output data.
    """

    weights = {
                "LT05_SR": {
                        "blue": 0.116,
                        "green": 0.010,
                        "red": 0.364,
                        "nir": 0.360,
                        "offset": 0.032,
                },
                "LE07_SR": {
                        "blue": 0.085,
                        "green": 0.057,
                        "red": 0.349,
                        "nir": 0.359,
                        "offset": 0.033,
                },
                "LC08_SR": {
                        "blue": 0.141,
                        "green": 0.114,
                        "red": 0.322,
                        "nir": 0.364,
                        "offset": 0.018,
                },
                "LC09_SR": {
                        "blue": 0.141,
                        "green": 0.114,
                        "red": 0.322,
                        "nir": 0.364,
                        "offset": 0.018,
                },
    }[product_name]

    reqs = ["blue", "green", "red", "nir"]
    if np.all([x in ds.data_vars for x in reqs]):
        ds["offset"] = xr.ones_like(ds["blue"])
        weights_da = xr.DataArray(data = list(weights.values()), 
                                coords = {"band": list(weights.keys())})
        ds["r0"] = ds[reqs + ["offset"]].to_array("band").weighted(weights_da).sum("band", skipna = False)
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in reqs if x not in ds.data_vars])}` is missing.")
    return ds

def calc_normalized_difference(ds, var, bands = ["nir", "red"]):
    """Calculate the normalized difference of two bands.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Name of the variable in which to store the normalized difference.
    bands : list, optional
        The two bands to use to calculate the norm. difference, by default ["nir", "red"].

    Returns
    -------
    xr.Dataset
        Output data.
    """
    if np.all([x in ds.data_vars for x in bands]):
        da = (ds[bands[0]] - ds[bands[1]]) / (ds[bands[0]] + ds[bands[1]])
        ds[var] = da.clip(-1, 1)
    else:
        log.warning(f"--> Couldn't calculate `{var}`, `{'` and `'.join([x for x in bands if x not in ds.data_vars])}` is missing.")
    return ds

def mask_uncertainty(ds, var, max_uncertainty = 3):
    if "lst_qa" in ds.data_vars:
        ds[var] = ds[var].where(ds["lst_qa"] <= max_uncertainty, np.nan)
    else:
        log.warning(f"--> Couldn't mask uncertain values.")
    return ds

def mask_invalid(ds, var):
    if "valid_range" in ds[var].attrs.keys():
        ds[var] = xr.where(
                            (ds[var] >= ds[var].valid_range[0]) & 
                            (ds[var] <= ds[var].valid_range[1]) &
                            (ds[var] != ds[var]._FillValue), 
                            ds[var], np.nan, keep_attrs=True)
    else:
        log.warning(f"--> Couldn't mask invalid values, since `valid_range` is not defined.")
    return ds

def apply_qa(ds, var, pixel_qa_flags = None, radsat_qa_flags = None, product_name = None):
    
    masks = list()

    if ("pixel_qa" in ds.data_vars) and not isinstance(pixel_qa_flags, type(None)):
        pixel_qa_bits = bitmasks.get_pixel_qa_bits(2, int(product_name.split("_")[0][-1]), 2)
        mask1 = bitmasks.get_mask(ds["pixel_qa"], pixel_qa_flags, pixel_qa_bits)
        masks.append(mask1)
        # Don't allow negative reflectances
        ds[var] = ds[var].clip(0, np.inf)

    if ("radsat_qa" in ds.data_vars) and not isinstance(radsat_qa_flags, type(None)):
        radsat_qa_bits = bitmasks.get_radsat_qa_bits(2, int(product_name.split("_")[0][-1]), 2)
        radsat_qa_flags = list(radsat_qa_bits.keys())
        mask2 = bitmasks.get_mask(ds["radsat_qa"], radsat_qa_flags, radsat_qa_bits)
        masks.append(mask2)

    if len(masks) >= 1:
        mask = np.invert(np.any(masks, axis = 0))
        ds[var] = ds[var].where(mask)

    return ds

def scale_data(ds, var):
    scale = getattr(ds[var], "scale_factor", 1.0)
    offset = getattr(ds[var], "add_offset", 0.0)
    ds[var] = ds[var] * scale + offset
    return ds

def default_vars(product_name, req_vars):
    # {x: [ds[x].dims, ds[x].long_name] for x in ds.data_vars}

    pixel_qa_flags_89 = ["dilated_cloud", "cirrus", "cloud", "cloud_shadow", "snow"]
    pixel_qa_flags_457 = ["dilated_cloud", "cloud", "cloud_shadow", "snow"]

    variables = {

        "LT05_SR": {
            'sr_band1':     [('YDim_sr_band1', 'XDim_sr_band1'), 'blue', [mask_invalid, scale_data, partial(apply_qa, product_name = "LT05_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band1"])]],
            'sr_band2':     [('YDim_sr_band2', 'XDim_sr_band2'), 'green', [mask_invalid, scale_data, partial(apply_qa, product_name = "LT05_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band2"])]],
            'sr_band3':     [('YDim_sr_band3', 'XDim_sr_band3'), 'red', [mask_invalid, scale_data, partial(apply_qa, product_name = "LT05_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band3"])]],
            'sr_band4':     [('YDim_sr_band4', 'XDim_sr_band4'), 'nir', [mask_invalid, scale_data, partial(apply_qa, product_name = "LT05_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band4"])]],
            'sr_band5':     [('YDim_sr_band5', 'XDim_sr_band5'), 'swir1', [mask_invalid, scale_data, partial(apply_qa, product_name = "LT05_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band5"])]],
            'sr_band7':     [('YDim_sr_band7', 'XDim_sr_band7'), 'swir2', [mask_invalid, scale_data, partial(apply_qa, product_name = "LT05_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band7"])]],
            'qa_pixel':     [('YDim_qa_pixel', 'XDim_qa_pixel'), 'pixel_qa', []],
            'qa_radsat':    [('YDim_qa_radsat', 'XDim_qa_radsat'), 'radsat_qa', []],
        },
        "LT05_ST": {
            'st_band6':     [('YDim_st_band6', 'XDim_st_band6'), 'lst', [mask_invalid, scale_data]],
            'st_qa':        [('YDim_st_qa', 'XDim_st_qa'), 'lst_qa', [mask_invalid, scale_data]],
        },
        "LE07_SR": {
            'sr_band1':     [('YDim_sr_band1', 'XDim_sr_band1'), 'blue', [mask_invalid, scale_data, partial(apply_qa, product_name = "LE07_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band1"])]],
            'sr_band2':     [('YDim_sr_band2', 'XDim_sr_band2'), 'green', [mask_invalid, scale_data, partial(apply_qa, product_name = "LE07_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band2"])]],
            'sr_band3':     [('YDim_sr_band3', 'XDim_sr_band3'), 'red', [mask_invalid, scale_data, partial(apply_qa, product_name = "LE07_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band3"])]],
            'sr_band4':     [('YDim_sr_band4', 'XDim_sr_band4'), 'nir', [mask_invalid, scale_data, partial(apply_qa, product_name = "LE07_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band4"])]],
            'sr_band5':     [('YDim_sr_band5', 'XDim_sr_band5'), 'swir1', [mask_invalid, scale_data, partial(apply_qa, product_name = "LE07_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band5"])]],
            'sr_band7':     [('YDim_sr_band7', 'XDim_sr_band7'), 'swir2', [mask_invalid, scale_data, partial(apply_qa, product_name = "LE07_SR", pixel_qa_flags = pixel_qa_flags_457, radsat_qa_flags = ["terrain_occlusion", "saturated_band7"])]],
            'qa_pixel':     [('YDim_qa_pixel', 'XDim_qa_pixel'), 'pixel_qa', []],
            'qa_radsat':    [('YDim_qa_radsat', 'XDim_qa_radsat'), 'radsat_qa', []],
        },
        "LE07_ST": {
            'st_band6':     [('YDim_st_band6', 'XDim_st_band6'), 'lst', [mask_invalid, scale_data]],
            'st_qa':        [('YDim_st_qa', 'XDim_st_qa'), 'lst_qa', [mask_invalid, scale_data]],
        },
        "LC08_SR": {
            'sr_band1':     [('YDim_sr_band1', 'XDim_sr_band1'), 'coastal',[mask_invalid, scale_data, partial(apply_qa, product_name = "LC08_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band1"])]],
            'sr_band2':     [('YDim_sr_band2', 'XDim_sr_band2'), 'blue', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC08_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band2"])]],
            'sr_band3':     [('YDim_sr_band3', 'XDim_sr_band3'), 'green', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC08_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band3"])]],
            'sr_band4':     [('YDim_sr_band4', 'XDim_sr_band4'), 'red', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC08_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band4"])]],
            'sr_band5':     [('YDim_sr_band5', 'XDim_sr_band5'), 'nir', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC08_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band5"])]],
            'sr_band6':     [('YDim_sr_band6', 'XDim_sr_band6'), 'swir1', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC08_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band6"])]],
            'sr_band7':     [('YDim_sr_band7', 'XDim_sr_band7'), 'swir2', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC08_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band7"])]],
            'qa_pixel':     [('YDim_qa_pixel', 'XDim_qa_pixel'), 'pixel_qa', []],
            'qa_radsat':    [('YDim_qa_radsat', 'XDim_qa_radsat'), 'radsat_qa', []],
        },
        "LC08_ST": {
            'st_band10':    [('YDim_st_band10', 'XDim_st_band10'), 'lst', [mask_invalid, scale_data]],
            'st_qa':        [('YDim_st_qa', 'XDim_st_qa'), 'lst_qa', [mask_invalid, scale_data]],
        },
        "LC09_SR": {
            'sr_band1':     [('YDim_sr_band1', 'XDim_sr_band1'), 'coastal',[mask_invalid, scale_data, partial(apply_qa, product_name = "LC09_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band1"])]],
            'sr_band2':     [('YDim_sr_band2', 'XDim_sr_band2'), 'blue', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC09_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band2"])]],
            'sr_band3':     [('YDim_sr_band3', 'XDim_sr_band3'), 'green', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC09_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band3"])]],
            'sr_band4':     [('YDim_sr_band4', 'XDim_sr_band4'), 'red', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC09_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band4"])]],
            'sr_band5':     [('YDim_sr_band5', 'XDim_sr_band5'), 'nir', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC09_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band5"])]],
            'sr_band6':     [('YDim_sr_band6', 'XDim_sr_band6'), 'swir1', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC09_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band6"])]],
            'sr_band7':     [('YDim_sr_band7', 'XDim_sr_band7'), 'swir2', [mask_invalid, scale_data, partial(apply_qa, product_name = "LC09_SR", pixel_qa_flags = pixel_qa_flags_89, radsat_qa_flags = ["terrain_occlusion", "saturated_band7"])]],
            'qa_pixel':     [('YDim_qa_pixel', 'XDim_qa_pixel'), 'pixel_qa', []],
            'qa_radsat':    [('YDim_qa_radsat', 'XDim_qa_radsat'), 'radsat_qa', []],
        },
        "LC09_ST": {
            'st_band10':    [('YDim_st_band10', 'XDim_st_band10'), 'lst', [mask_invalid, scale_data]],
            'st_qa':        [('YDim_st_qa', 'XDim_st_qa'), 'lst_qa', [mask_invalid, scale_data]],
        }

    }

    req_dl_vars = {

        "LT05_SR": {
            'blue': ['sr_band1', 'qa_pixel', 'qa_radsat'],
            'green': ['sr_band2', 'qa_pixel', 'qa_radsat'],
            'red': ['sr_band3', 'qa_pixel', 'qa_radsat'],
            'nir': ['sr_band4', 'qa_pixel', 'qa_radsat'],
            'swir1': ['sr_band5', 'qa_pixel', 'qa_radsat'],
            'swir2': ['sr_band7', 'qa_pixel', 'qa_radsat'],
            'pixel_qa': ['qa_pixel'],
            'radsat_qa': ['qa_radsat'],
            'ndvi': ['sr_band3', 'sr_band4', 'qa_pixel', 'qa_radsat'],
            'r0': ['sr_band1', 'sr_band2', 'sr_band3', 'sr_band4', 'qa_pixel', 'qa_radsat'],
        },
        "LT05_ST": {
            'lst': ['st_band6', 'st_qa'],
            'lst_qa': ['st_qa'],
        },
        "LE07_SR": {
            'blue': ['sr_band1', 'qa_pixel', 'qa_radsat'],
            'green': ['sr_band2', 'qa_pixel', 'qa_radsat'],
            'red': ['sr_band3', 'qa_pixel', 'qa_radsat'],
            'nir': ['sr_band4', 'qa_pixel', 'qa_radsat'],
            'swir1': ['sr_band5', 'qa_pixel', 'qa_radsat'],
            'swir2': ['sr_band7', 'qa_pixel', 'qa_radsat'],
            'pixel_qa': ['qa_pixel'],
            'radsat_qa': ['qa_radsat'],
            'ndvi': ['sr_band3', 'sr_band4', 'qa_pixel', 'qa_radsat'],
            'r0': ['sr_band1', 'sr_band2', 'sr_band3', 'sr_band4', 'qa_pixel', 'qa_radsat'],
        },
        "LE07_ST": {
            'lst': ['st_band6', 'st_qa'],
            'lst_qa': ['st_qa'],
        },
        "LC08_SR": {
            'coastal': ['sr_band1', 'qa_pixel', 'qa_radsat'],
            'blue': ['sr_band2', 'qa_pixel', 'qa_radsat'],
            'green': ['sr_band3', 'qa_pixel', 'qa_radsat'],
            'red': ['sr_band4', 'qa_pixel', 'qa_radsat'],
            'nir': ['sr_band5', 'qa_pixel', 'qa_radsat'],
            'swir1': ['sr_band6', 'qa_pixel', 'qa_radsat'],
            'swir2': ['sr_band7', 'qa_pixel', 'qa_radsat'],
            'pixel_qa': ['qa_pixel'],
            'radsat_qa': ['qa_radsat'],
            'ndvi': ['sr_band4', 'sr_band5', 'qa_pixel', 'qa_radsat'],
            'r0': ['sr_band2', 'sr_band3', 'sr_band4', 'sr_band5', 'qa_pixel', 'qa_radsat'],
        },
        "LC08_ST": {
            'lst': ['st_band10', 'st_qa'],
            'lst_qa': ['st_qa'],
        },
        "LC09_SR": {
            'coastal': ['sr_band1', 'qa_pixel', 'qa_radsat'],
            'blue': ['sr_band2', 'qa_pixel', 'qa_radsat'],
            'green': ['sr_band3', 'qa_pixel', 'qa_radsat'],
            'red': ['sr_band4', 'qa_pixel', 'qa_radsat'],
            'nir': ['sr_band5', 'qa_pixel', 'qa_radsat'],
            'swir1': ['sr_band6', 'qa_pixel', 'qa_radsat'],
            'swir2': ['sr_band7', 'qa_pixel', 'qa_radsat'],
            'pixel_qa': ['qa_pixel'],
            'radsat_qa': ['qa_radsat'],
            'ndvi': ['sr_band4', 'sr_band5', 'qa_pixel', 'qa_radsat'],
            'r0': ['sr_band2', 'sr_band3', 'sr_band4', 'sr_band5', 'qa_pixel', 'qa_radsat'],
        },
        "LC09_ST": {
            'lst': ['st_band10', 'st_qa'],
            'lst_qa': ['st_qa'],
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def default_post_processors(product_name, req_vars):

    post_processors = {
        "LT05_SR": {
            'coastal': [],
            'blue': [],
            'green': [],
            'red': [],
            'nir': [],
            'swir1': [],
            'swir2': [],
            'pixel_qa': [],
            'radsat_qa': [],
            'ndvi': [calc_normalized_difference],
            'r0': [partial(calc_r0, product_name = "LT05_SR")],
            },
        "LT05_ST": {
            'lst': [mask_uncertainty],
            'lst_qa': [],
        },
        "LE07_SR": {
            'coastal': [],
            'blue': [],
            'green': [],
            'red': [],
            'nir': [],
            'swir1': [],
            'swir2': [],
            'pixel_qa': [],
            'radsat_qa': [],
            'ndvi': [calc_normalized_difference],
            'r0': [partial(calc_r0, product_name = "LE07_SR")],
            },
        "LE07_ST": {
            'lst': [mask_uncertainty, ],
            'lst_qa': [],
        },
        "LC08_SR": {
            'coastal': [],
            'blue': [],
            'green': [],
            'red': [],
            'nir': [],
            'swir1': [],
            'swir2': [],
            'pixel_qa': [],
            'radsat_qa': [],
            'ndvi': [calc_normalized_difference],
            'r0': [partial(calc_r0, product_name = "LC08_SR")],
            },
        "LC08_ST": {
            'lst': [mask_uncertainty],
            'lst_qa': [],
        },
        "LC09_SR": {
            'coastal': [],
            'blue': [],
            'green': [],
            'red': [],
            'nir': [],
            'swir1': [],
            'swir2': [],
            'pixel_qa': [],
            'radsat_qa': [],
            'ndvi': [calc_normalized_difference],
            'r0': [partial(calc_r0, product_name = "LC09_SR")],
            },
        "LC09_ST": {
            'lst': [mask_uncertainty],
            'lst_qa': [],
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def espa_api(endpoint, verb='get', body=None, uauth=None):
    """ Suggested simple way to interact with the ESPA JSON REST API """
    # auth_tup = uauth if uauth else (username, password)
    host = 'https://espa.cr.usgs.gov/api/v1/'
    response = getattr(r, verb)(host + endpoint, auth=uauth, json=body)
    data = response.json()
    if isinstance(data, dict):
        messages = data.pop("messages", None)  
        if messages:
            log.info(json.dumps(messages, indent=4))
    try:
        response.raise_for_status()
    except Exception as e:
        log.warning(str(e))
        return None
    else:
        return data

def search_stac(latlim, lonlim, timelim, product_name, extra_search_kwargs):

    timelim = adjust_timelim_dtype(timelim)
    sd = datetime.datetime.strftime(timelim[0], "%Y-%m-%dT00:00:00Z")
    ed = datetime.datetime.strftime(timelim[1], "%Y-%m-%dT23:59:59Z")
    search_dates = f"{sd}/{ed}"

    bb = [lonlim[0], latlim[0], lonlim[1], latlim[1]]

    # TODO this doesnt work, so filtering manually later, doc example: `'platform': {'or':['LANDSAT_8','LANDSAT_9']}`.
    # platform = {
    #             "LC08_SR": "LANDSAT_8",
    #             "LC09_SR": "LANDSAT_9",
    #             "LE07_SR": "LANDSAT_7",
    #             "LT05_SR": "LANDSAT_5",

    #             "LC08_ST": "LANDSAT_8",
    #             "LC09_ST": "LANDSAT_9",
    #             "LE07_ST": "LANDSAT_7",
    #             "LT05_ST": "LANDSAT_5",
    #             }[product_name]
    # search_kwargs = {   
    #                     **{'platform': {'or':[platform]}},
    #                     **extra_search_kwargs
    #                 }
    search_kwargs = extra_search_kwargs

    stac = 'https://landsatlook.usgs.gov/stac-server' # Landsat STAC API Endpoint
    stac_response = r.get(stac).json() 
    catalog_links = stac_response['links']
    search = [l['href'] for l in catalog_links if l['rel'] == 'search'][0]   #retreive search endpoint from STAC Catalog

    params = dict()
    params['collections'] = ['landsat-c2l2-sr','landsat-c2l2-st']
    params['limit'] = 1000
    params['bbox'] = bb
    params['datetime'] = search_dates
    params['query'] = search_kwargs

    query = r.post(search, json=params, ).json()   # send POST request to the stac-search endpoint with params passed in

    ids = np.unique([x["id"].replace("_ST", "").replace("_SR", "") for x in query["features"] if product_name.split("_")[0] in x["id"]]).tolist()

    log.info(f"--> Found {len(ids)} scenes.")

    return ids, query

def request_scenes(ids, image_extents):

    uauth = pywapor.collect.accounts.get("EARTHEXPLORER") 

    order = espa_api('available-products', 
                        body = {"inputs": list(ids)}, 
                        uauth = uauth)
    


    if "not implemented" in order.keys():
        missing = order.pop("not implemented")
        log.warning(f"--> Some scenes could not be found (`{'`, `'.join(missing)}`).")

    req_prods = ["l1"]
    for sensor in order.keys():
        order[sensor]["products"] = [x for x in req_prods if x in order[sensor]["products"]]

    order['projection'] = {"lonlat": None}
    order['format'] = 'netcdf'
    order['resampling_method'] = 'nn'
    order['note'] = f'pyWaPOR_{pywapor.__version__}'
    order["image_extents"] = image_extents

    log.info(f"--> Placing order for {len(ids)} scenes.")

    order_response = espa_api('order', verb='post', body = order, uauth = uauth)

    time.sleep(10)

    return order_response

def unpack(fp, folder):
    fn, _ = os.path.splitext(os.path.split(fp)[-1])
    subfolder = os.path.join(folder, fn)
    if not os.path.exists(subfolder):
        os.makedirs(subfolder)
    shutil.unpack_archive(fp, subfolder)

def check_availabilty(product_folder, product_name, scene_ids):
    paths = glob.glob(os.path.join(product_folder, "**", "*.nc"), recursive = True)
    regex_str = r'_.{4}_\d{6}_\d{8}_\d{8}_\d{2}_T\d.nc'
    unpacked_scenes = set([os.path.splitext(os.path.split(f)[-1])[0] for f in paths if re.search(f'{product_name.split("_")[0]}{regex_str}', f)])
    packed_scenes = {[y.name for y in tarfile.open(x, encoding='utf-8').getmembers() if ".nc" in y.name][0][:-3]: x for x in glob.glob(os.path.join(product_folder, "**", f"{product_name.split('_')[0]}*.tar.gz"), recursive = True)}
    to_unpack = [fp for k, fp in packed_scenes.items() if (k in scene_ids) and (k not in unpacked_scenes)]
    for fp in to_unpack:
        unpack(fp, product_folder)
    available_scenes = unpacked_scenes.union(set(packed_scenes.keys())).intersection(scene_ids)
    return available_scenes

def update_order_statuses(scene_ids, product_folder, product_name, to_download, to_wait, to_request, uauth, image_extents, verbose = True):

    all_orders = espa_api(f"item-status", uauth = uauth)
    available_scenes = check_availabilty(product_folder, product_name, scene_ids)
    
    for order_id, order in all_orders.items():

        # Get specific order details.
        order_folder = os.path.join(product_folder, "orders")
        order_details_fp = os.path.join(order_folder, f"{order_id}.json")
        if os.path.isfile(order_details_fp):
            with open(order_details_fp, "r") as fp:
                order_details = json.load(fp)
        else:
            order_details = espa_api(f"order/{order_id}", uauth = uauth)
            with open(order_details_fp, "w") as fp:
                json.dump(order_details , fp) 

        if order_details["product_opts"].get("image_extents") != image_extents:
            continue

        for scene in order:

            if (scene["status"] == "complete") and (scene["name"] in scene_ids) and (scene["name"] not in available_scenes):
                to_download[scene["name"]] = scene["product_dload_url"]
                to_request.discard(scene["name"])
                to_wait.discard(scene["name"])
            elif (scene["status"] == "oncache") and (scene["name"] in scene_ids) and (scene["name"] not in available_scenes):
                to_request.discard(scene["name"])
                to_wait.add(scene["name"])
            elif (scene["status"] == "onorder") and (scene["name"] in scene_ids) and (scene["name"] not in available_scenes):
                to_request.discard(scene["name"])
                to_wait.add(scene["name"])
            elif (scene["status"] == "queued") and (scene["name"] in scene_ids) and (scene["name"] not in available_scenes):
                to_request.discard(scene["name"])
                to_wait.add(scene["name"])
            elif (scene["status"] == "processing") and (scene["name"] in scene_ids) and (scene["name"] not in available_scenes):
                to_request.discard(scene["name"])
                to_wait.add(scene["name"])
            elif (scene["status"] == "error") and (scene["name"] in scene_ids) and (scene["name"] not in available_scenes):
                to_wait.discard(scene["name"])
                to_request.discard(scene["name"])
                if not verbose:
                    log.info(f"--> Error in `{scene['name']}` request, `{scene['note']}`.")
            elif (scene["status"] == "retry") and (scene["name"] in scene_ids) and (scene["name"] not in available_scenes):
                to_request.discard(scene["name"])
                to_wait.add(scene["name"])
            elif (scene["status"] == "unavailable") and (scene["name"] in scene_ids) and (scene["name"] not in available_scenes):
                to_request.discard(scene["name"])
                to_wait.discard(scene["name"])
                if not verbose:
                    log.info(f"--> Scene `{scene['name']}` is unavailable, `{scene['note']}`.")
            elif (scene["status"] == "cancelled") and (scene["name"] in scene_ids) and (scene["name"] not in available_scenes):
                to_request.add(scene["name"])
                to_wait.discard(scene["name"])
                if not verbose:
                    log.info(f"Scene `{scene['name']}` order is cancelled, `{scene['note']}`. Ordering again.")
            elif (scene["name"] in scene_ids) and (scene["name"] not in available_scenes):
                to_request.add(scene["name"])
                to_wait.discard(scene["name"])
            else:
                ...

    return available_scenes, to_download, to_wait, to_request

def download_scenes(scene_ids, product_folder, product_name, latlim, lonlim, max_attempts = 5, wait_time = 300):

    image_extents = {
            "east": lonlim[1],
            "north": latlim[1],
            "south": latlim[0],
            "west": lonlim[0],
            "units": "dd",
        }

    uauth = pywapor.collect.accounts.get("EARTHEXPLORER")

    attempt = 0

    # Check which scenes already exist locally
    available_scenes = check_availabilty(product_folder, product_name, scene_ids)
    to_request = set(scene_ids).difference(available_scenes)
    to_wait = set()
    to_download = dict()

    while len(to_download) + len(to_wait) + len(to_request) > 0 and attempt < max_attempts:

        if attempt > 0:
            log.info(f"--> Waiting {wait_time} seconds before trying again.")
            time.sleep(wait_time)

        # Update statutes
        available_scenes, to_download, to_wait, to_request = update_order_statuses(scene_ids, product_folder, product_name, to_download, to_wait, to_request, uauth, image_extents, verbose = True)
        # log.info(f"--> {len(available_scenes)} scenes available, {len(to_download)} ready for download, {len(to_request)} need to be requested and {len(to_wait)} are being processed.")

        # Request missing scenes on (ESPA)
        if len(to_request) > 0:
            _ = request_scenes(to_request, image_extents)

        # Download completed scenes (ESPA)
        if len(to_download) > 0:
            log.info(f"--> Downloading {len(to_download)} scenes.")
            fps = download_urls(list(to_download.values()), product_folder)
            for fp in fps:
                unpack(fp, product_folder)
            to_download = dict()

        available_scenes, to_download, to_wait, to_request = update_order_statuses(scene_ids, product_folder, product_name, to_download, to_wait, to_request, uauth, image_extents, verbose = False)
        # log.info(f"--> {len(available_scenes)} scenes available, {len(to_download)} ready for download, {len(to_request)} need to be requested and {len(to_wait)} are being processed.")

        attempt += 1
        log.info(f"--> {len(available_scenes)} scenes collected in attempt {attempt}/{max_attempts}, waiting for {len(to_wait) + len(to_request)} more scenes.")
        
    return available_scenes, to_download, to_request, to_wait

def _process_scene(scene, product_folder, product_name, variables, example_ds = None):

    ds = xr.open_dataset(scene, mask_and_scale=False, chunks = "auto")[list(variables.keys())]

    if len(set([ds[x].shape for x in ds.data_vars])) > 1:
        log.warning("--> Not all variables have identical shapes.")

    xdim = ds[list(variables.values())[0][0][1]].values
    ydim = ds[list(variables.values())[0][0][0]].values

    renames1 = {k: v[1] for k, v in variables.items()}
    renames2 = {v[0][0]: "y" for v in variables.values()}
    renames3 = {v[0][1]: "x" for v in variables.values()}

    ds = ds.rename_dims({**renames2, **renames3})
    ds = ds.drop(list(ds.coords.keys())).assign_coords({"x": xdim, "y": ydim})
    ds = ds.rename(renames1)

    crs = rasterio.crs.CRS.from_epsg(4326)
    ds = ds.rio.write_crs(crs)
    ds = ds.rio.write_grid_mapping("spatial_ref")

    ds = ds.sortby("y", ascending = False)
    ds = ds.sortby("x")

    mtl_fp = glob.glob(os.path.join(product_folder, "**", f"*{ds.LPGSMetadataFile.replace('.txt', '.xml')}*"), recursive=True)[0]
    with open(mtl_fp,"r") as f:
        xml_content = f.read()
    mtl = xmltodict.parse(xml_content)

    # Clip and pad to bounding-box
    if isinstance(example_ds, type(None)):
        example_ds = make_example_ds(ds, product_folder, crs, bb = None, example_ds_fp = os.path.join(product_folder, f"example_ds_{product_name}.nc"))#[lonlim[0], latlim[0], lonlim[1], latlim[1]])
    ds = ds.rio.reproject_match(example_ds).chunk("auto")
    ds = ds.assign_coords({"x": example_ds.x, "y": example_ds.y})

    # Add time dimension to data arrays.
    ds = ds.expand_dims({"time": 1})

    # Set the correct time.
    date_str = mtl['LANDSAT_METADATA_FILE']["IMAGE_ATTRIBUTES"]["DATE_ACQUIRED"]
    time_str = mtl['LANDSAT_METADATA_FILE']["IMAGE_ATTRIBUTES"]["SCENE_CENTER_TIME"]
    datetime_str = date_str + " " + time_str.replace("Z", "")
    ds = ds.assign_coords({"time":[np.datetime64(datetime_str)]})

    # Apply variable specific functions.
    for vars in variables.values():
        for func in vars[2]:
            ds, _ = apply_enhancer(ds, vars[1], func)

    # Cleanup attributes
    for var in ["x", "y", "time"]:
        ds[var].attrs = {}

    return ds, example_ds

def process_scenes(fp, scene_paths, product_folder, product_name, variables, post_processors):
    
    dss = list()

    log.info(f"--> Processing {len(scene_paths)} scenes.").add()

    chunks = {"time": 1, "x": 1000, "y": 1000}

    example_ds = None
    for i, scene in enumerate(scene_paths):
        ds, example_ds = _process_scene(scene, product_folder, product_name, variables, example_ds = example_ds)
        fp_temp = scene.replace(".nc", "_temp.nc")
        ds = save_ds(ds, fp_temp, chunks = chunks, encoding = "initiate", label = f"({i+1}/{len(scene_paths)}) Processing `{os.path.split(scene)[-1]}`.")
        dss.append(ds)
    remove_ds(example_ds)

    ds = xr.concat(dss, "time")
    
    # Apply general product functions.
    for var, funcs in post_processors.items():
        for func in funcs:
            ds, label = apply_enhancer(ds, var, func)
            log.info(label)

    # Remove unrequested variables.
    ds = ds[list(post_processors.keys())]
    
    for var in ds.data_vars:
        ds[var].attrs = {}

    ds = ds.sortby("time")

    # Save final netcdf.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="invalid value encountered in true_divide")
        warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")
        ds = save_ds(ds, fp, chunks = "auto", encoding = "initiate", label = f"Merging files.")

    # Remove intermediate files.
    for x in dss:
        remove_ds(x)

    return ds

def download(folder, latlim, lonlim, timelim, product_name, 
                req_vars, variables = None, post_processors = None, 
                extra_search_kwargs = {'eo:cloud_cover': {'gte': 0, 'lt': 30}},
                max_attempts = 24, wait_time = 300):
    """Order, Download and preprocess Landsat scenes.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.
    extra_search_kwargs : dict, optional
        Extra keywords passed to the scene searching API, by default {'eo:cloud_cover': {'gte': 0, 'lt': 30}}
    max_attempts : int, optional
        Maximum number of retries, by default 24.
    wait_time : int, optional
        Wait time in seconds between retries, by default 300.

    Returns
    -------
    xr.Dataset
        Dataset with the requested variables.

    Raises
    ------
    ValueError
        Raised when not all found scenes could be downloaded within the max_attempts.
    """


    adjust_logger(True, folder, "INFO")

    product_folder = os.path.join(folder, "LANDSAT")
    order_folder = os.path.join(product_folder, "orders")

    if not os.path.exists(product_folder):
        os.makedirs(product_folder)

    if not os.path.exists(order_folder):
        os.makedirs(order_folder)

    fn = os.path.join(product_folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    # Search scene IDs (STAC)
    scene_ids, _ = search_stac(latlim, lonlim, timelim, product_name, extra_search_kwargs)

    if len(scene_ids) == 0:
        return None

    # Order and download scenes (ESPA)
    available_scenes, to_download, to_request, to_wait = download_scenes(scene_ids, product_folder, product_name, latlim, lonlim, max_attempts = max_attempts, wait_time = wait_time)
    # log.info(f"available_scenes = {available_scenes}")

    # log.info(f"available_scenes: {len(available_scenes)}, scene_ids: {len(scene_ids)}, to_wait: {len(to_wait)}, to_download: {len(to_download)}, to_request: {len(to_request)} \n\n {available_scenes} \n {scene_ids} \n {to_wait}")
    if len(available_scenes) < len(scene_ids) and len(to_wait) > 0:
        raise ValueError(f"Waiting for order of {len(to_wait)} scenes to finish.", len(to_wait))

    # Process scenes.
    scene_paths = [glob.glob(os.path.join(product_folder, "**", f"*{x}.nc"), recursive = True)[0] for x in available_scenes]
    # log.info(f"scene_paths = {scene_paths}")
    ds = process_scenes(fn, scene_paths, product_folder, product_name, variables, post_processors)

    return ds[req_vars_orig]

# if __name__ == "__main__":

#     # tests = {
#     #     # "LT05_SR": ["2010-03-29", "2010-04-25"], 
#     #     # "LE07_SR": ["2010-03-29", "2010-04-25"], 
#     #     # "LC08_SR": ["2022-03-29", "2022-04-25"], 
#     #     # "LC09_SR": ["2022-03-29", "2022-04-25"]
#     #     "LC08_ST": ["2022-03-29", "2022-04-25"],
#     # }

#     sources = "level_3"
#     period = 3
#     area = "pakistan_south"

#     lonlim, latlim = {
#         "fayoum":           ([30.2,  31.2],  [28.9,  29.7]),
#         "pakistan_south":   ([67.70, 67.90], [26.35, 26.55]),
#         "pakistan_hydera":  ([68.35, 68.71], [25.49, 25.73]),
#     }[area]

#     timelim = {
#         0: [datetime.date(2019, 10, 1), datetime.date(2019, 10, 11)],
#         1: [datetime.date(2022, 5, 1), datetime.date(2022, 5, 11)],
#         2: [datetime.date(2022, 10, 1), datetime.date(2022, 10, 11)],
#         3: [datetime.date(2022, 8, 1), datetime.date(2022, 10, 1)],
#         4: [datetime.date(2021, 8, 1), datetime.date(2021, 10, 1)],
#     }[period]

#     folder = f"/Users/hmcoerver/Local/{area}_{sources}_{period}" #

#     adjust_logger(True, folder, "INFO")

#     bin_length = "DEKAD"

#     from pywapor.general import compositer
#     bins = compositer.time_bins(timelim, bin_length)

#     adjusted_timelim = [bins[0], bins[-1]]
#     timelim = [adjusted_timelim[0] - np.timedelta64(3, "D"), 
#                         adjusted_timelim[1] + np.timedelta64(3, "D")]

#     product_name = "LE07_SR"

#     variables = None
#     post_processors = None
#     extra_search_kwargs = {'eo:cloud_cover': {'gte': 0, 'lt': 30}}
#     max_attempts = 24
#     wait_time = 300
    # folder = f"/Users/hmcoerver/Local/landsat_test2"
    # adjust_logger(True, folder, "INFO")
    # for product_name, timelim in tests.items():
    #     print(product_name, timelim)
        
    #     latlim = [28.9, 29.7]
    #     lonlim = [30.2, 31.2]
    #     # timelim = ["2022-03-29", "2022-04-25"]
    #     # product_name = "LC08"
    #     req_vars = ["lst"]
    #     # req_vars = ["r0"]
    #     variables = None
    #     post_processors = None
    #     # example_ds = None
    #     extra_search_kwargs = {'eo:cloud_cover': {'gte': 0, 'lt': 30}}
    #     ds = download(folder, latlim, lonlim, timelim, product_name, 
    #                     req_vars, variables = variables, post_processors = post_processors, 
    #                     extra_search_kwargs = extra_search_kwargs)




import xarray as xr
import numpy as np
from pywapor.general.processing_functions import save_ds, open_ds
from pywapor.enhancers.smooth.plotters import make_overview
from pywapor.general.logger import log
from pywapor.enhancers.smooth.core import _wt1, _wt2, cve1, second_order_diff_matrix, dist_to_finite

def xr_dist_to_finite(y, dim = "time"):

    if not dim in y.dims:
        raise ValueError

    out = xr.apply_ufunc(
        dist_to_finite, y, y[dim],
        input_core_dims=[[dim], [dim]],
        output_core_dims=[[dim]],
        vectorize=False,
        dask="parallelized",
    )

    return out

def xr_choose_func(y, lmbdas, dim):
    
    funcs = [_wt1, _wt2]
    y_dims = getattr(y, "ndim", 0)
    lmbdas_dims = getattr(lmbdas, "ndim", 0)
    if y_dims in [2,3] and lmbdas_dims in [1]:
        wt_func = funcs[1]
        icd = [[dim],[],["lmbda"],[],[],[],[],[]]
        ocd = [["lmbda", dim]]
    elif y_dims in [2] and lmbdas_dims in [2]:
        raise ValueError
    else:
        wt_func = funcs[0]
        icd = [[dim],[],[],[],[],[],[],[]]
        ocd = [[dim]]

    return wt_func, icd, ocd

def xr_cve(y, x, lmbdas, u):

    # Check dimension and dtypes are valid.
    x, y, lmbdas, dim = assert_dtypes(x, y, lmbdas)

    # Normalize x-coordinates
    x = (x - x.min()) / (x.max() - x.min()) * x.size

    # Create x-aware delta matrix.
    A = second_order_diff_matrix(x)

    # Make default u weights if necessary.
    if isinstance(u, type(None)):
        u = np.ones(x.shape)

    # Apply whittaker smoothing along axis.
    cves = xr.apply_ufunc(
        cve1, lmbdas, y, A, u,
        input_core_dims = [["lmbda"],[dim],[],[]],
        output_core_dims = [["lmbda"]],
        dask = "allowed",
        )
    
    return cves

def xr_wt(y, x, lmbdas, u = None, a = 0.5, min_drange = -np.inf, 
          max_drange = np.inf, max_iter = 10):

    # Check dimension and dtypes are valid.
    x, y, lmbdas, dim = assert_dtypes(x, y, lmbdas)

    # Normalize x-coordinates
    x = (x - x.min()) / (x.max() - x.min()) * x.size

    # Create x-aware delta matrix.
    A = second_order_diff_matrix(x)

    # Make default u weights if necessary.
    if isinstance(u, type(None)):
        u = np.ones(x.shape)

    # Choose which vectorized function to use.
    _wt, icd, ocd = xr_choose_func(y, lmbdas, dim)

    # Make sure lmbdas is chunked similar to y.
    if not isinstance(y.chunk, type(None)):
        lmbdas = lmbdas.chunk({k: v for k, v in y.unify_chunks().chunksizes.items() if k in lmbdas.dims})

    # Apply whittaker smoothing along axis.
    z = xr.apply_ufunc(
        _wt, y, A, lmbdas, u, a, min_drange, max_drange, max_iter,
        input_core_dims = icd,
        output_core_dims = ocd,
        dask = "allowed",
        )

    # Add some metadata.
    z.attrs = {"a": f"{a:.2f}", 
               "min_drange": str(min_drange), 
               "max_drange": str(max_drange)}

    return z

def make_weights(sensor_da, weights):
    weights_dict = {{v: k for k, v in sensor_da.attrs.items()}[sensor]: weight for sensor, weight in weights.items() if sensor in sensor_da.attrs.values()}
    values = np.array(list(weights_dict.values()))
    coords = np.array(list(weights_dict.keys()), dtype = int)
    transformer = xr.DataArray(values, dims=["sensor"], coords = {"sensor": coords})
    u = transformer.sel(sensor = sensor_da).values
    return u

def assert_dtypes(x, y, lmbdas):

    # Check x and y.
    assert x.ndim == 1
    if isinstance(x, xr.DataArray) and isinstance(y, xr.DataArray):
        dim_name = x.dims[0]
        assert dim_name in y.dims
    elif isinstance(x, np.ndarray) and isinstance(y, xr.DataArray):
        dim_names = [k for k,v in y.sizes.items() if v == x.size]
        if len(dim_names) != 1:
            raise ValueError
        else:
            dim_name = dim_names[0]
            x = xr.DataArray(x, dims = [dim_name], coords = y.dim_name)
    elif isinstance(x, xr.DataArray) and isinstance(y, np.ndarray):
        x = x.values
        dim_name = None
        assert x.size == y.shape[-1]
    elif isinstance(x, np.ndarray) and isinstance(y, np.ndarray):
        assert x.size == y.shape[-1]
        dim_name = None
    else:
        raise TypeError

    # Check lmbdas.
    if isinstance(lmbdas, float) or isinstance(lmbdas, int) or isinstance(lmbdas, list):
        lmbdas = np.array(lmbdas)
    assert lmbdas.ndim <= 2
    if isinstance(x, xr.DataArray) and (isinstance(lmbdas, np.ndarray) or np.isscalar(lmbdas)):
        if not np.isscalar(lmbdas):
            assert lmbdas.ndim <= 1
            if lmbdas.ndim == 0:
                lmbdas = float(lmbdas)
            else:
                lmbdas = xr.DataArray(lmbdas, dims = ["lmbda"], coords = {"lmbda": lmbdas})
        # else:
        lmbdas = xr.DataArray(lmbdas)
    elif isinstance(x, xr.DataArray) and isinstance(lmbdas, xr.DataArray):
        ...
    elif isinstance(x, np.ndarray) and (isinstance(lmbdas, np.ndarray) or np.isscalar(lmbdas)):
        if lmbdas.ndim == 0:
            lmbdas = float(lmbdas)
        elif lmbdas.ndim == 2 and y.ndim == 3:
            assert y.shape[:-1] == lmbdas.shape
    elif isinstance(x, np.ndarray) and isinstance(lmbdas, xr.DataArray):
        lmbdas = lmbdas.values
        if lmbdas.ndim == 0:
            lmbdas = float(lmbdas)
    else:
        raise TypeError
    
    return x, y, lmbdas, dim_name

def whittaker_smoothing(ds, var, lmbdas = 100., weights = None, a = 0.5, 
                        max_iter = 10, out_fh = None, xdim = "time", max_dist = None,
                        new_x = None, export_all = False, chunks = {}, make_plots = None,
                        valid_drange = [-np.inf, np.inf], **kwargs):
    """Apply Whittaker smoothing to a variable in a dataset.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing a variable `var` which has at least the dimension called `xdim`.
    var : str
        Name of the variable in `ds` to smooth. The variable can have a shape
        (nt), (nx,ny,nt) or (n,nt) with nt the size of the dimension `xdim`.
    lmbdas : int | float | np.array | xr.DataArray, optional
        Lambda value to use for smoothing, shape should be (), (m) or (nx, ny), by default 100.
    a : float, optional
        Apply quantile smoothing, value can be between 0.01 and 0.99. When 0.5,
        no iterations (limited by `max_iter`) are done, by default 0.5
    max_iter : int, optional
        Maximum number of iterations to perform when applying quantile smoothing, by default 10
    out_fh : _type_, optional
        Path to store results, when None the output is saved in the same folder as the input, by default None.
    xdim : str, optional
        The dimension describing the x values of the variable to be smoothed, by default "time"
    new_x : _type_, optional
        Extra values to add to the x dimension (used for interpolation), by default None
    export_all : bool, optional
        Whether to save only the smoothed data or also diagnostics information, by default False

    Returns
    -------
    xr.Dataset
        Dataset containing at least (depending on `export_all`) a variable called `{var}_smoothed`.
    """
    # Check if the chunk overwriter is correct (i.e. doesnt chunk along core-dims)
    if not getattr(chunks, "get", lambda x,y: False)(xdim, False) == -1 and (not chunks == {}):
        log.warning(f"--> Adjusting defined chunks (`{chunks}`), to avoid chunks along core dimension ({xdim}).")
        chunks = {dim: {True: -1, False: getattr(chunks, "get", lambda x,y: chunks)(dim, "auto")}[dim == xdim] for dim in ds.dims}

    # Check if the ds is chunked correctly in case `chunks` is not set.
    is_chunked = ds.chunksizes.get(xdim, False)
    if is_chunked and (getattr(is_chunked, "__len__", lambda: 1)() != 1) and (chunks == {}):
        chunks = {dim: {True: -1, False: "auto"}[dim == xdim] for dim in ds.dims}
        log.warning(f"--> Adjusting chunks to (`{chunks}`), to avoid chunks along core dimension ({xdim}).")

    # Chunk dataset.
    ds = ds.chunk(chunks)

    # Add new x values.
    if not isinstance(new_x, type(None)) and getattr(xdim, '__len__', lambda: 0)() > 0:
        ds = xr.concat([ds, xr.Dataset({xdim: new_x})], dim = xdim).drop_duplicates(xdim).chunk(chunks).sortby(xdim).chunk(chunks)
        if "sensor" in ds.data_vars:
            sensor_id = np.nanmax(np.unique(ds["sensor"])) + 1
            ds["sensor"] = ds["sensor"].fillna(sensor_id).assign_attrs({str(int(sensor_id)): "Interp."})
            if not isinstance(weights, type(None)):
                weights["Interp."] = 0.0

    # Add lmbdas as coordinate.
    ds = ds.assign_coords({"lmbda": {True: np.array([lmbdas]), False: lmbdas}[np.isscalar(lmbdas)]})

    # Create weights.
    if not isinstance(weights, type(None)) and ("sensor" in ds.data_vars):
        u = make_weights(ds["sensor"], weights)
        source_legend = {i: f"{x} [{weights.get(x, np.nan):.2f}]" for i, x in ds["sensor"].attrs.items()}
        ds["sensor"].attrs = source_legend
    else:
        u = None

    # Only do this when more then one lmbda is provided.
    if np.any([getattr(lmbdas, 'size', 1) > 1, export_all, "plot_folder" in kwargs.keys()]):
        if a != 0.5 and getattr(lmbdas, 'size', 1) > 1:
            log.warning(f"--> Picking lambda when `a` != 0.5 (`a` = {a}) is unsupported and can result in unexpected behaviour.")
        if not np.isinf(valid_drange[0]) and getattr(lmbdas, 'size', 1) > 1:
            log.warning(f"--> Picking lambda with forces bounds (min = {valid_drange[0]}) is unsupported and can result in unexpected behaviour.")
        if not np.isinf(valid_drange[1]) and getattr(lmbdas, 'size', 1) > 1:
            log.warning(f"--> Picking lambda with forces bounds (max = {valid_drange[1]}) is unsupported and can result in unexpected behaviour.")
        ds["cves"] = xr_cve(ds[var], ds[xdim], ds["lmbda"], u)
        ds["lmbda_sel"] = ds["cves"].idxmin("lmbda")
        lmbdas = ds["lmbda_sel"]

    # Smooth the data.
    ds[f"{var}_smoothed"] = xr_wt(ds[var], ds[xdim], lmbdas = lmbdas, u = u, a = a,
                                      min_drange = valid_drange[0], max_drange = valid_drange[1],
                                      max_iter = max_iter)

    # Interpolate any data that was skipped because of stability issues.
    ds[f"{var}_smoothed"] = ds[f"{var}_smoothed"].interpolate_na(dim = xdim)

    # Mask values that are too far away from any measurement.
    if not isinstance(max_dist, type(None)):
        xdist = xr_dist_to_finite(ds[var], dim = xdim)
        ds[f"{var}_smoothed"] = ds[f"{var}_smoothed"].where(xdist <= max_dist, drop = False)

    # Create a plot
    if "plot_folder" in kwargs.keys():
        ds = ds.compute()
        # NOTE kwargs are passed on to make_overview.
        # kwargs = {"point_method": "equally_spaced", "n": 3, "offset": 0.1, "folder": r""}
        # kwargs = {"point_method": "worst", "n": 5, "xdim": "time", "folder": r""}
        plot_folder = kwargs.pop("plot_folder")
        make_overview(ds, var, plot_folder, **kwargs)

    # Drop irrelevant data.
    if not export_all:
        ds = ds[[f"{var}_smoothed"]]
        if not isinstance(new_x, type(None)) and getattr(xdim, '__len__', lambda: 0)() > 0:
            ds = ds.sel({xdim: new_x})

    if not isinstance(out_fh, type(None)):
        ds = save_ds(ds, out_fh, encoding = "initiate", label = f"Applying whittaker smoothing ({var}).")

    # Give warnings if data is outside of defined range.
    if not np.isinf(valid_drange[0]):
        min_bound = float(ds[f"{var}_smoothed"].min().values)
        if min_bound < valid_drange[0]:
            log.warning(f"--> Minimum of `{var}_smoothed` is smaller than `valid_drange` min ({min_bound:.2f} < {valid_drange[0]}).")

    if not np.isinf(valid_drange[1]):
        max_bound = float(ds[f"{var}_smoothed"].max().values)
        if max_bound > valid_drange[1]:
            log.warning(f"--> Maximum of `{var}_smoothed` is larger than `valid_drange` max ({max_bound:.2f} > {valid_drange[1]}).")

    return ds

if __name__ == "__main__":

    import matplotlib.pyplot as plt

    ds = open_ds(r"/Users/hmcoerver/Desktop/test.nc")[["ndvi"]].drop_vars("lmbda")

    var = "ndvi"

    lmbdas = 100.
    weights = None
    a = 0.5
    max_iter = 10
    out_fh = None
    xdim = "time"
    max_dist = None
    new_x = None
    export_all = False
    chunks = {}
    valid_drange = [-np.inf, np.inf]



import functools
import os
import numpy as np
from itertools import product
import xarray as xr
import pywapor
from pywapor.collect.protocol import cog
import copy
from pywapor.general.processing_functions import open_ds, save_ds, remove_ds
from pywapor.enhancers.apply_enhancers import apply_enhancer
from pywapor.enhancers.dem import calc_slope, calc_aspect
from pywapor.general.logger import log

def tiles_intersect(latlim, lonlim, product_name):
    """Creates a list of server-side filenames for tiles that intersect with `latlim` and
    `lonlim` for the selected product. 

    Parameters
    ----------
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    product_name : str
        Name of the product to download.

    Returns
    -------
    list
        Server-side filenames for tiles.
    """
    with open(os.path.join(pywapor.collect.__path__[0], f"product/{product_name}.txt")) as src:
        all_tiles = src.read().split("\n")
    we_tiles = np.arange(np.floor(lonlim[0]), np.ceil(lonlim[1]), 1).astype(int).tolist()
    ns_tiles = np.arange(np.floor(latlim[0]), np.ceil(latlim[1]), 1).astype(int).tolist()
    tiles = list(product(we_tiles, ns_tiles))
    dl_tiles = list()
    res = {"GLO30": 10, "GLO90": 30}[product_name]
    for tile in tiles:
        we_sign = {-1: "W", 1: "E"}[np.sign(tile[0])]
        ns_sign = {-1: "S", 1: "N"}[np.sign(tile[1])]
        fn = f"Copernicus_DSM_COG_{res}_{ns_sign}{tile[1]:02}_00_{we_sign}{tile[0]:03}_00_DEM"
        if fn in all_tiles:
            dl_tiles.append(fn)
    return dl_tiles

def url_func(product_name, fn):
    """Returns a url at which to collect COPERNICUS data.

    Parameters
    ----------
    product_name : str
        Name of the product to download.
    fn : str
        Name of the server-side filename to download.

    Returns
    -------
    str
        The url.
    """
    url = {
            "GLO30": f"/vsis3/copernicus-dem-30m/{fn}/{fn}.tif",
            "GLO90": f"/vsis3/copernicus-dem-90m/{fn}/{fn}.tif",
            }[product_name]
    return url

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {
        "GLO30": {
                "Band1": [("lat", "lon"), "z"],
                "crs": [(), "spatial_ref"],
                    },
        "GLO90": {
                "Band1": [("lat", "lon"), "z"],
                "crs": [(), "spatial_ref"],
                    },                
    }

    req_dl_vars = {
        "GLO30": {
            "z": ["Band1", "crs"],
            "slope": ["Band1", "crs"],
            "aspect": ["Band1", "crs"],
        },
        "GLO90": {
            "z": ["Band1", "crs"],
            "slope": ["Band1", "crs"],
            "aspect": ["Band1", "crs"],
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def default_post_processors(product_name, req_vars = ["z"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["z"].

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "GLO30": {
            "z": [],
            "slope": [calc_slope],
            "aspect": [calc_aspect],
        },
        "GLO90": {
            "z": [],
            "slope": [calc_slope],
            "aspect": [calc_aspect],
        },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def download(folder, latlim, lonlim, product_name = "GLO30", req_vars = ["z"],
                variables = None, post_processors = None, **kwargs):
    """Download COPERNICUS data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str, optional
        Name of the product to download, by default "GLO30".
    req_vars : list, optional
        Which variables to download for the selected product, by default ["z"].
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """

    folder = os.path.join(folder, "COPERNICUS")

    final_fp = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(final_fp):
        existing_ds = open_ds(final_fp)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]
            
    spatial_buffer = True
    if spatial_buffer:
        dx = dy = 0.00027778
        latlim = [latlim[0] - dy, latlim[1] + dy]
        lonlim = [lonlim[0] - dx, lonlim[1] + dx]

    dl_tiles = tiles_intersect(latlim, lonlim, product_name)

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    coords = {"x": ("lon", lonlim), "y": ("lat", latlim)}

    gdal_config_options = {'AWS_REGION': 'eu-central-1', 'AWS_NO_SIGN_REQUEST': ''}

    dss = list()

    log.info(f"--> Downloading {len(dl_tiles)} {product_name} tiles.").add()

    for fn in dl_tiles:

        url_func_part = functools.partial(url_func, fn = fn)
        fp = os.path.join(folder, f"{fn}.nc")

        if os.path.isfile(fp):
            ds = open_ds(fp)
        else:
            ds = cog.download(fp, product_name, coords, variables, {}, url_func_part, gdal_config_options = gdal_config_options, waitbar = False)

        dss.append(ds)

    log.sub()

    ds = xr.concat(dss, "stack").median("stack")

    # Apply product specific functions.
    for var, funcs in post_processors.items():
        for func in funcs:
            ds, label = apply_enhancer(ds, var, func)
            log.info(label)
    
    # Remove unrequested variables.
    ds = ds[list(post_processors.keys())]
    
    # Save final output.
    ds = save_ds(ds, final_fp, encoding = "initiate", label = f"Saving {product_name}.nc")

    for nc in dss:
        remove_ds(nc)

    return ds[req_vars_orig]

if __name__ == "__main__":

    folder = r"/Users/hmcoerver/Local/cog_test"
    product_name = r"GLO90" # r"GLO90" r"GLO30
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    req_vars = ["z", "slope", "aspect"]
    variables = None
    post_processors = None

    ds = download(folder, latlim, lonlim, product_name = product_name, req_vars = req_vars,
                    variables = variables, post_processors = post_processors)


import numpy as np
from pywapor.et_look_v2_v3 import constants as c
import xarray as xr
import warnings

def initial_sensible_heat_flux_canopy_daily(rn_24_canopy, t_24_init):
    r"""
    Computes the initial sensible heat flux before the iteration which solves
    the stability corrections. The first estimation of transpiration is used
    to estimate the initial sensible heat flux.

    .. math ::

        H_{canopy}=Q_{canopy}^{*}-T

    Parameters
    ----------
    rn_24_canopy : float
        daily net radiation for the canopy, 
        :math:`Q_{canopy}^{*}`
        [W m-2]
    t_24_init : float
        initial estimate of daily transpiration, 
        :math:`T`
        [W m-2]

    Returns
    -------
    h_canopy_24_init : float
        initial estimate of the sensible heat flux, 
        :math:`H^{canopy}`
        [W m-2]
    """

    return rn_24_canopy - t_24_init


def initial_sensible_heat_flux_soil_daily(rn_24_soil, e_24_init, g0_24):
    r"""
    Computes the initial sensible heat flux before the iteration which solves
    the stability corrections. The first estimation of transpiration is used
    to estimate the initial sensible heat flux.

    .. math ::

        H_{soil}=Q_{soil}^{*}-G_{0}-E

    Parameters
    ----------
    rn_24_soil : float
        daily net radiation for the soil, 
        :math:`Q_{canopy}^{*}`
        [W m-2]
    g0_24 : float
        daily soil heat flux, 
        :math:`G_{0}`
        [W m-2]
    e_24_init : float
        initial estimate of daily evaporation, 
        :math:`E`
        [W m-2]

    Returns
    -------
    h_soil_24_init : float
        initial estimate of the sensible heat flux, 
        :math:`H_{canopy}`
        [W m-2]
    """
    return rn_24_soil - g0_24 - e_24_init


def initial_friction_velocity_daily(u_b_24, z0m, disp, z_b=100):
    r"""
    Computes the initial friction velocity without using stability corrections.

    .. math ::

        u_{*}=\frac{k \cdot u_{b}}{ln\left(\frac{z_{b}-d}{z_{0,m}}\right)}

    Parameters
    ----------
    u_b_24 : float
        daily wind speed at blending heigt, 
        :math:`u_{b}`
        [m s-1]
    z0m : float
        surface roughness, 
        :math:`z_{0,m}`
        [m]
    disp : float
        displacement height, 
        :math:`d`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]


    Returns
    -------
    u_star_24_init : float
        initial estimate of the daily friction velocity, 
        :math:`u_{*}`
        [m s-1]
    """
    return (c.k * u_b_24) / (np.log((z_b - disp) / z0m))


def initial_friction_velocity_soil_daily(u_b_24, disp, z_b=100):
    r"""
    Computes the initial firction velocity without using stability corrections.

    .. math ::

        u_{*}=\frac{k \cdot u_{b}}{ln\left(\frac{z_{b}-d}{z_{0,m}}\right)}

    Parameters
    ----------
    u_b_24 : float
        daily wind speed at blending heigt, 
        :math:`u_{b}`
        [m s-1]
    disp : float
        displacement height, 
        :math:`d`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]


    Returns
    -------
    u_star_24_soil_init : float
        initial estimate of the daily friction velocity for soil, 
        :math:`u_{*}`
        [m s-1]
    """
    return (c.k * u_b_24) / (np.log((z_b - disp) / c.z0_soil))


def monin_obukhov_length(h_flux, ad, u_star, t_air_k):
    r"""
    Computes the Monin-Obukhov length. The Monin-Obukhov length is used to
    describe the effects of buoyancy on turbulent flows. The Monin-Obukhov
    length is usually negative during daytime.

    .. math ::

        L=\frac{-\rho c_{p}u_{*}^{3}T_{a}}{kgH_{canopy}}

    Parameters
    ----------
    h_flux : float
        sensible heat flux, 
        :math:`H`
        [W m-2]
    ad : float
        air density, 
        :math:`\rho`
        [kg m-3]
    u_star : float
        Monin Obukhov length, 
        :math:`L`
        [m]
    t_air_k : float
        air tempererature in kelvin, 
        :math:`T_{a}`
        [K]

    Returns
    -------
    monin : float
        monin obukhov length, 
        :math:`L`
        [m]
    """
    return (-ad * c.sh * u_star ** 3 * t_air_k) / (c.k * c.g * h_flux)


def stability_parameter(monin, disp, z_b=100):
    r"""
    Computes the stability parameter introduced by Monin and Obukhov. This
    parameter includes effects of both shear stress and buoyancy on turbulence.
    It is applicable to blending height.

    .. math ::

        x_{b}=1-16\left(\frac{z_{b}-d}{L}\right)^{0.25}

    Parameters
    ----------
    monin : float
        monin obukhov length, 
        :math:`L`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]
    disp : float
        displacement height, 
        :math:`d`
        [m]

    Returns
    -------
    x_b : float
        stability parameter used in stability correction, 
        :math:`x_{b}`
        [-]
    """
    return (1 - 16 * ((z_b - disp) / monin)) ** 0.25


def stability_factor(x_b):
    r"""
    Computes the stability correction for heat at blending height.

    .. math ::

        \psi_{h,b}=2 \cdot \ln\left(\frac{1+x_{b}}{2}\right)+
        \ln\left(\frac{1+x_{b}^{2}}{2}\right)-
        2 \cdot \arctan\left(x_{b}\right)+0.5 \cdot \pi

    Parameters
    ----------
    x_b : float
        stability parameter used in stability correction, 
        :math:`x_{b}`
        [-]

    Returns
    -------
    sf : float
        stability correction for heat, 
        :math:`\psi_{h,b}`
        [-]
    """
    return (
        2 * np.log((1 + x_b) / 2)
        + np.log((1 + x_b ** 2) / 2)
        - 2 * np.arctan(x_b)
        + 0.5 * np.pi
    )


def stability_parameter_obs(monin, z_obs):
    r"""
    Computes the stability parameter introduced by Monin and Obukhov. This
    parameter includes effects of both shear stress and buoyancy on turbulence.
    It is applicable to observation height.

    .. math ::

        x_{obs}=1-16\left(\frac{z_{obs}}{L}\right)^{0.25}

    Parameters
    ----------
    monin : float
        monin obukhov length, 
        :math:`L`
        [m]
    z_obs : float
        observation height, 
        :math:`z_{obs}`
        [m]

    Returns
    -------
    x_b_obs : float
        stability parameter used in stability correction for observation height, 
        :math:`x_{obs}`
        [-]
    """
    return (1 - 16 * (z_obs / monin)) ** 0.25


def stability_correction_heat_obs(x_b_obs):
    r"""
    Computes the stability correction for heat at observation height.

    .. math ::

        \psi_{h,obs}=2 \cdot \ln\left(\frac{1+x_{obs}^{2}}{2}\right)

    Parameters
    ----------
    x_b_obs : float
        stability parameter used in stability correction for observation height, 
        :math:`x_{obs}`
        [-]

    Returns
    -------
    sf_obs : float
        stability correction for heat for observation height, 
        :math:`\psi_{h,obs}`
        [-]
    """
    return 2 * np.log((1 + x_b_obs ** 2) / 2)


def friction_velocity(u_b, z_b, z0m, disp, sf):
    r"""
    Computes the friction velocity.

    .. math ::

        u_{*}=\frac{k \cdot u_{b}}{ln\left(\frac{z_{b}-d}{z_{0,m}}\right)-\psi_{h,b}}

    Parameters
    ----------
    u_b : float
        windspeed at blending height, 
        :math:`u_{b}`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]
    z0m : float
        roughness length, 
        :math:`z_{0,m}`
        [m]
    disp : float
        displacement height, 
        :math:`d`
        [m]
    sf : float
        stability factor at blending height, 
        :math:`\psi_{h,b}`
        [m]

    Returns
    -------
    u_star : float
        friction velocity, 
        :math:`u_{*}`
        [m s-1]
    """
    return (c.k * u_b) / (np.log((z_b - disp) / z0m) - sf)


def ra_canopy(
    h_canopy_init, t_air_k, u_star_init, ad, z0m, disp, u_b, z_obs=2, z_b=100, iter_ra=3
):
    r"""
    Computes the aerodynamical resistance for canopy using an iterative
    approach. The iteration is needed to compute the frication velocity at
    blending height Iteration stops either after five iterations or
    if the difference between two subsequent estimations is less than 0.01.

    .. math ::
    
        L &= \frac{-\rho c_{p}u_{*}^{3}T_{a}}{kgH_{canopy}}\\
        x_{b} &= 1-16\left(\frac{z_{b}-d}{L}\right)^{0.25}\\
        \psi_{h,b} &= 2 \ln\left(\frac{1+z_{b}}{2}\right)+
        \ln\left(\frac{1+z_{b}^{2}}{2}\right)-
        2\arctan\left(x_{b}\right)+0.5\pi\\
        u_{*} &= \frac{ku_{b}}{ln\left(\frac{z_{b}-d}{z_{0,m}}\right)-\psi_{h,b}}


    The friction velocity is independent of height. So this value can be used
    to calculate together with the stability correction for heat on observation
    heigth the aerodynamical resistance.

    .. math ::

        x_{obs}&=1-16 \cdot \left(\frac{z_{obs}}{L}\right)^{0.25} \\
        \psi_{h,obs}&=2 \cdot \ln\left(\frac{1+x_{obs}^{2}}{2}\right) \\
        r_{a,canopy}&=\frac{\ln\left(\frac{z_{obs}-d}
        {0.1 \cdot z_{0,m}}\right)-\psi_{h,obs}}{k \cdot u_{*}}

    Parameters
    ----------
    h_canopy_init : float
        initial estimate of the sensible heat flux, 
        :math:`H^{canopy}`
        [W m-2]
    t_air_k : float
        air tempererature in kelvin, 
        :math:`T_{a}`
        [K]
    u_star_init : float
        initial estimate of the daily friction velocity, 
        :math:`u_{*}`
        [m s-1]
    ad : float
        air density, 
        :math:`\rho`
        [kg m-3]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]
    z_obs : float
        observation height, 
        :math:`z_{obs}`
        [m]
    z0m : float
        roughness length, 
        :math:`z_{0,m}`
        [m]
    disp : float
        displacement height, 
        :math:`d`
        [m]
    u_b : float
        windspeed at blending height, 
        :math:`u_{b}`
        [m/s]
    iter_ra : integer
        number of iterations for aerodynamical resistance, 
        :math:`n_{ra}`
        [-]

    Returns
    -------
    ra_canopy : float
        aerodynamical resistance for canopy, 
        :math:`r_{a,canopy}`
        [s m-1]
    """
    h_flux = h_canopy_init
    u_star_start = u_star_init
    iteration = 0
    epsilon = 10.0
    while (iteration < iter_ra) and (np.nanmax(epsilon) > 0.01):
        iteration += 1
        monin = monin_obukhov_length(h_flux, ad, u_star_start, t_air_k)
        if isinstance(monin, xr.DataArray):
            x_b = xr.where(monin > 0, 1, stability_parameter(monin, disp, z_b))
        else:
            x_b = np.where(monin > 0, 1, stability_parameter(monin, disp, z_b))
        sf = stability_factor(x_b)
        u_star = friction_velocity(u_b, z_b, z0m, disp, sf)
        epsilon = abs(u_star - u_star_start)
        u_star_start = u_star

    if isinstance(monin, xr.DataArray):
        x_b_obs = xr.where(monin <= 0, stability_parameter_obs(monin, z_obs), 1)
        sf_obs = xr.where(monin <= 0, stability_correction_heat_obs(x_b_obs), 0)
    else:
        x_b_obs = np.where(monin <= 0, stability_parameter_obs(monin, z_obs), 1)
        sf_obs = np.where(monin <= 0, stability_correction_heat_obs(x_b_obs), 0)

    disp = np.minimum(disp, 1.5)
    ra = (np.log((z_obs - disp) / (0.1 * z0m)) - sf_obs) / (c.k * u_star)
    
    if isinstance(monin, xr.DataArray):
        ra = ra.clip(25, 500)
    else:
        ra = np.clip(ra, 25, 500)

    return ra


def transpiration(
    rn_24_canopy,
    ssvp_24,
    ad_24,
    vpd_24,
    psy_24,
    r_canopy,
    h_canopy_24_init,
    t_air_k_24,
    u_star_24_init,
    z0m,
    disp,
    u_b_24,
    z_obs=2,
    z_b=100,
    iter_h=3,
):
    r"""
    Computes the transpiration using an iterative approach. The iteration is
    needed to compute the aerodynamical resistance.Iteration stops either after
    five iterations orif the difference between two subsequent estimations
    is less than 0.01. The iteration is started with an estimate on :math:`H`
    using the initial guess without stability corrections. Subsequent
    iterations use the guess with stability corrections.

    .. math ::

        T=\frac{\Delta\left(Q_{canopy}^{*}\right)+\rho c_{p}\
           frac{\Delta_{e}}{r_{a,canopy}}}{\Delta+
           \gamma\left(1+\frac{r_{canopy}}{r_{a,canopy}}\right)}

    Parameters
    ----------
    rn_24_canopy : float
        net radiation for the canopy, 
        :math:`Q^{*}_{canopy}`
        [Wm-2]
    ssvp_24 : float
       daily slope of saturated vapour pressure curve
       :math:`\Delta_{24}`
       [mbar K-1]
    ad_24 : float
        daily air density, 
        :math:`\rho_{24}`
        [kg m-3]
    vpd_24 : float
       daily vapour pressure deficit
       :math:`\Delta_{e,24}`
       [mbar]
    psy_24 : float
        daily psychrometric constant, 
        :math:`\gamma_{24}`
        [mbar K-1]
    r_canopy : float
        canopy resistance, 
        :math:`r_{canopy}`
        [sm-1]
    h_canopy_24_init : float
        initial estimate of the sensible heat flux, 
        :math:`H^{canopy}`
        [W m-2]
    t_air_k_24 : float
        daily air tempererature in kelvin, 
        :math:`T_{a}`
        [K]
    u_star_24_init : float
        initial estimate of the daily friction velocity, 
        :math:`u_{*}`
        [m s-1]
    z0m : float
        roughness length, 
        :math:`z_{0,m}`
        [m]
    disp : float
        displacement height, 
        :math:`d`
        [m]
    u_b_24 : float
        daily windspeed at blending height, 
        :math:`u_{b}`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]
    z_obs : float
        observation height, 
        :math:`z_{obs}`
        [m]
    iter_h : integer
        number of iterations for sensible heat flux, 
        :math:`n_h`
        [-]

    Returns
    -------
    t_24 : float
        daily transpiration energy equivalent, 
        :math:`T_{24}`
        [W m-2]
    """
    iteration = 0
    epsilon = 10.0
    h_start = h_canopy_24_init

    with warnings.catch_warnings():

        warnings.filterwarnings("ignore", message="invalid value encountered in power")
        warnings.filterwarnings("ignore", message="invalid value encountered in true_divide")
        warnings.filterwarnings("ignore", message="invalid value encountered in divide")
        warnings.filterwarnings("ignore", message="divide by zero encountered in log")
        warnings.filterwarnings("ignore", message="divide by zero encountered in true_divide")
        warnings.filterwarnings("ignore", message="divide by zero encountered in divide")

        while (iteration < iter_h) and (np.nanmax(epsilon) > 0.01):
            iteration += 1
            
            ra_canopy_start = ra_canopy(
                h_start, t_air_k_24, u_star_24_init, ad_24, z0m, disp, u_b_24, z_obs, z_b
            )

            t = (ssvp_24 * rn_24_canopy + ad_24 * c.sh * (vpd_24 / ra_canopy_start)) / (
                ssvp_24 + psy_24 * (1 + r_canopy / ra_canopy_start)
            )
            h = rn_24_canopy - t
            epsilon = abs(h - h_start)
            h_start = h

    return t


def ra_soil(
    h_soil_24_init, t_air_k, u_star_24_init, ad, disp, u_b, z_obs=2, z_b=100, iter_ra=3
):
    r"""
    Computes the aerodynamical resistance for canopy using an iterative
    approach. The iteration is needed to compute the friction velocity at
    blending height Iteration stops either after five iterations or
    if the difference between two subsequent estimations is less than 0.01.

    .. math ::

        L &= \frac{-\rho c_{p}u_{*}^{3}T_{a}}{kgH_{soil}}\\
        x_{b} &= 1-16\left(\frac{z_{b}-d}{L}\right)^{0.25}\\
        \psi_{h,b} &= 2\ln\left(\frac{1+z_{b}}{2}\right)+
        \ln\left(\frac{1+z_{b}^{2}}{2}\right)-
        2\arctan\left(x_{b}\right)+0.5\pi\\
        u_{*} &= \frac{ku_{b}}{ln\left(\frac{z_{b}-d}{z_{0,soil}}\right)
        -\psi_{h,b}}


    The friction velocity is independent of height. So this value can be used
    to calculate together with the stability correction for heat on observation
    heigth the aerodynamical resistance.

    .. math ::

        x_{obs} &= 1-16\left(\frac{z_{obs}}{L}\right)^{0.25} \\
        \psi_{h,obs} &= 2\ln\left(\frac{1+x_{obs}^{2}}{2}\right) \\
        r_{a,soil} &= \frac{\ln\left(\frac{z_{obs}-d}
        {0.1z_{0,soil}}\right)-\psi_{h,obs}}{ku_{*}}

    Parameters
    ----------
    h_soil_24_init : float
        initial estimate of the sensible heat flux for soil, 
        :math:`H^{soil}`
        [W m-2]
    t_air_k : float
        air tempererature in kelvin, 
        :math:`T_{a}`
        [K]
    u_star_24_init : float
        initial estimate of the daily friction velocity, 
        :math:`u_{*}`
        [m s-1]
    ad : float
        air density, 
        :math:`\rho`
        [kg m-3]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]
    z_obs : float
        observation height, 
        :math:`z_{obs}`
        [m]
    disp : float
        displacement height, 
        :math:`d`
        [m]
    u_b : float
        windspeed at blending height, 
        :math:`u_{b}`
        [m]
    iter_ra : integer
        number of iterations for aerodynamical resistance, 
        :math:`n_{ra}`
        [-]

    Returns
    -------
    ra_soil : float
        aerodynamical resistance for soil, 
        :math:`r_{a,soil}`
        [s m-1]
    """
    h_flux = h_soil_24_init
    u_star_start = u_star_24_init
    iteration = 0
    epsilon = 10
    while (iteration < iter_ra) and (np.nanmax(epsilon) > 0.01):
        iteration += 1
        monin = monin_obukhov_length(h_flux, ad, u_star_start, t_air_k)
        if isinstance(monin, xr.DataArray):
            x_b = xr.where(monin > 0, 0, stability_parameter(monin, disp, z_b))
        else:
            x_b = np.where(monin > 0, 0, stability_parameter(monin, disp, z_b))
        sf = stability_factor(x_b)
        u_star = friction_velocity(u_b, z_b, c.z0_soil, disp, sf)
        epsilon = abs(u_star - u_star_start)
        u_star_start = u_star

    if isinstance(monin, xr.DataArray):
        x_b_obs = xr.where(monin <= 0, stability_parameter_obs(monin, z_obs), 1)
        sf_obs = xr.where(monin <= 0, stability_correction_heat_obs(x_b_obs), 0)
    else:
        x_b_obs = np.where(monin <= 0, stability_parameter_obs(monin, z_obs), 1)
        sf_obs = np.where(monin <= 0, stability_correction_heat_obs(x_b_obs), 0)

    # 1.5 limit is from ETLook IDL
    disp = np.minimum(disp, 1.5)
    ra = (np.log((z_obs - disp) / (0.1 * c.z0_soil)) - sf_obs) / (c.k * u_star)
    ra = np.maximum(ra, 25)

    return ra


def evaporation(
    rn_24_soil,
    g0_24,
    ssvp_24,
    ad_24,
    vpd_24,
    psy_24,
    r_soil,
    h_soil_24_init,
    t_air_k_24,
    u_star_24_soil_init,
    disp,
    u_b_24,
    z_b=100,
    z_obs=2,
    iter_h=3,
):
    r"""
    Computes the evaporation using an iterative approach. The iteration is
    needed to compute the aerodynamic resistance.Iteration stops either after
    five iterations or if the difference between two subsequent estimations
    is less than 0.01. The iteration is started with an estimate on :math:`H`
    using the initial guess without stability corrections. Subsequent
    iterations use the guess with stability corrections.

    .. math ::

        E=\frac{\Delta\left(Q_{soil}^{*}-G\right)+
          \rho c_{p}\frac{\Delta_{e}}{r_{a,soil}}}
          {\Delta+\gamma\left(1+\frac{r_{soil}}{r_{a,soil}}\right)}

    Parameters
    ----------
    rn_24_soil : float
        net radiation for the soil, 
        :math:`Q^{*}_{canopy}`
        [Wm-2]
    g0_24 : float
        daily soil heat flux, 
        :math:`G`
        [Wm-2]
    ssvp_24 : float
       daily slope of saturated vapour pressure curve
       :math:`\Delta_{24}`
       [mbar K-1]
    ad_24 : float
        daily air density, 
        :math:`\rho_{24}`
        [kg m-3]
    vpd_24 : float
       daily vapour pressure deficit
       :math:`\Delta_{e,24}`
       [mbar]
    psy_24 : float
        daily psychrometric constant, 
        :math:`\gamma_{24}`
        [mbar K-1]
    r_soil : float
        soil resistance, 
        :math:`r_{soil}`
        [sm-1]
    h_soil_24_init : float
        initial estimate of the sensible heat flux for soil, 
        :math:`H^{soil}`
        [W m-2]
    t_air_k_24 : float
        daily air temperature in kelvin, 
        :math:`T_{a}`
        [K]
    u_star_24_soil_init : float
        initial estimate of the daily friction velocity for soil, 
        :math:`u_{*}`
        [m s-1]
    disp : float
        displacement height, 
        :math:`d`
        [m]
    u_b_24 : float
        daily wind speed at blending height, 
        :math:`u_{b}`
        [m]
    z_b : float
        blending height, 
        :math:`z_{b}`
        [m]
    z_obs : float
        observation height, 
        :math:`z_{obs}`
        [m]
    iter_h : integer
        number of iterations for sensible heat flux, 
        :math:`n_h`
        [-]

    Returns
    -------
    e_24 : float
        daily evaporation energy equivalent, 
        :math:`E_{24}`
        [W m-2]
    """
    iteration = 0
    epsilon = 10
    h_start = h_soil_24_init

    with warnings.catch_warnings():

        warnings.filterwarnings("ignore", message="invalid value encountered in power")
        warnings.filterwarnings("ignore", message="invalid value encountered in divide")
        warnings.filterwarnings("ignore", message="invalid value encountered in true_divide")
        warnings.filterwarnings("ignore", message="divide by zero encountered in log")
        warnings.filterwarnings("ignore", message="divide by zero encountered in power")
        warnings.filterwarnings("ignore", message="divide by zero encountered in divide")
        
        while (iteration < iter_h) and (np.nanmax(epsilon) > 0.1):
            iteration += 1
            ra_soil_start = ra_soil(
                h_start, t_air_k_24, u_star_24_soil_init, ad_24, disp, u_b_24, z_obs, z_b
            )
            e = (
                ssvp_24 * (rn_24_soil - g0_24) + ad_24 * c.sh * (vpd_24 / ra_soil_start)
            ) / (ssvp_24 + psy_24 * (1 + r_soil / ra_soil_start))
            h = rn_24_soil - g0_24 - e
            epsilon = abs(h - h_start)
            h_start = h

    return e


def transpiration_mm(t_24, lh_24):
    r"""
    Computes the canopy transpiration based on the Penman Monteith equation
    adapted for canopy.

    .. math ::

        T=T \cdot \frac{d_{sec}}{\lambda_{24}}

    where the following constants are used

    * :math:`d_{sec}` seconds in the day = 86400 [s]

    Parameters
    ----------
    t_24 : float
        daily transpiration energy equivalent, 
        :math:`E^{0}`
        [W m-2]
    lh_24 : float
        daily latent heat of evaporation, 
        :math:`\lambda_{24}`
        [J/kg]

    Returns
    -------
    t_24_mm : float
        daily transpiration in mm, 
        :math:`T`
        [mm d-1]
    """
    return t_24 * c.day_sec / lh_24


def evaporation_mm(e_24, lh_24):
    r"""
    Computes the soil evaporation based on the Penman Monteith equation
    adapted for soils.

    .. math ::

        E=E \cdot \frac{d_{sec}}{\lambda_{24}}

    where the following constants are used

    * :math:`d_{sec}` seconds in the day = 86400 [s]

    Parameters
    ----------
    e_24 : float
        daily evaporation energy equivalent, 
        :math:`E^{0}`
        [W m-2]
    lh_24 : float
        daily latent heat of evaporation, 
        :math:`\lambda_{24}`
        [J/kg]

    Returns
    -------
    e_24_mm : float
        daily evaporation in mm, 
        :math:`E`
        [mm d-1]
    """
    return e_24 * c.day_sec / lh_24




"""
https://opendap.cr.usgs.gov/opendap/hyrax/SRTMGL1_NUMNC.003/contents.html
"""
import datetime
import os
import json
import pywapor.collect
import xarray as xr
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds
import pywapor.collect.accounts as accounts
from shapely.geometry.polygon import Polygon
from shapely.geometry import shape
import copy
import pywapor.collect.protocol.opendap as opendap
from pywapor.enhancers.dem import calc_slope, calc_aspect
import numpy as np

def tiles_intersect(latlim, lonlim):
    """Creates a list of server-side filenames for tiles that intersect with `latlim` and
    `lonlim` for the selected product. 

    Parameters
    ----------
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.

    Returns
    -------
    list
        Server-side filenames for tiles.
    """
    with open(os.path.join(pywapor.collect.__path__[0], "product/SRTM30_tiles.geojson")) as f:
        features = json.load(f)["features"]
    aoi = Polygon.from_bounds(lonlim[0], latlim[0], lonlim[1], latlim[1])
    tiles = list()
    for feature in features:
        shp = shape(feature["geometry"])
        tile = feature["properties"]["dataFile"]
        if shp.intersects(aoi):
            tiles.append(tile.split(".")[0])
    return tiles

def default_vars(product_name, req_vars = ["z"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    variables = {
        "30M": {
            "SRTMGL1_DEM": [("time", "lat", "lon"), "z"],
            "crs": [(), "spatial_ref"],
            }
    }

    req_dl_vars = {
        "30M": {
            "z": ["SRTMGL1_DEM", "crs"],
            "slope": ["SRTMGL1_DEM", "crs"],
            "aspect": ["SRTMGL1_DEM", "crs"],
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def drop_time(ds):
    return ds.isel(time=0).drop("time")

def default_post_processors(product_name, req_vars = ["z"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """
    
    post_processors = {
        "30M": {
            "z": [],
            "aspect": [calc_aspect],
            "slope": [calc_slope],
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def fn_func(product_name, tile):
    """Returns a client-side filename at which to store data.

    Parameters
    ----------
    product_name : str
        Name of the product to download.
    tile : str
        Name of the server-side tile to download.

    Returns
    -------
    str
        Filename.
    """
    fn = f"{product_name}_{tile}.nc"
    return fn

def url_func(product_name, tile):
    """Returns a url at which to collect MERRA2 data.

    Parameters
    ----------
    product_name : None
        Not used.
    tile : str
        Name of the server-side tile to download.

    Returns
    -------
    str
        The url.
    """
    url = f"https://opendap.cr.usgs.gov/opendap/hyrax/SRTMGL1_NC.003/{tile}.SRTMGL1_NC.ncml.nc4?"
    return url

def download(folder, latlim, lonlim, product_name = "30M", req_vars = ["z"], variables = None, post_processors = None, **kwargs):
    """Download SRTM data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str, optional
        Name of the product to download, by default "30M".
    req_vars : list, optional
        Which variables to download for the selected product, by default ["z"].
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "SRTM")
    
    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    spatial_buffer = True
    if spatial_buffer:
        dx = dy = 0.0002777777777777768
        latlim = [latlim[0] - dy, latlim[1] + dy]
        lonlim = [lonlim[0] - dx, lonlim[1] + dx]

    timelim = [datetime.date(2000, 2, 10), datetime.date(2000, 2, 12)]
    tiles = tiles_intersect(latlim, lonlim)
        
    coords = {"x": ["lon", lonlim], "y": ["lat", latlim], "t": ["time", timelim]}
    
    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    data_source_crs = None
    parallel = False
    spatial_tiles = True
    un_pw = accounts.get("NASA")
    request_dims = True

    ds = opendap.download(fn, product_name, coords, 
                variables, post_processors, fn_func, url_func, un_pw = un_pw, 
                tiles = tiles, data_source_crs = data_source_crs, parallel = parallel, 
                spatial_tiles = spatial_tiles, request_dims = request_dims)

    return ds[req_vars_orig]

if __name__ == "__main__":

    folder = r"/Users/hmcoerver/Local/cog2_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 1), datetime.date(2020, 7, 11)]
    req_vars = ["z", "aspect"]

    # fn = os.path.join(os.path.join(folder, "SRTM"), "30M.nc")
    # if os.path.isfile(fn):
    #     os.remove(fn)

    # # SRTM.
    ds = download(folder, latlim, lonlim, req_vars = req_vars)
    print(ds.rio.crs, ds.rio.grid_mapping)

